import pytest
from quam.core.operation.operation import Operation
from quam.core.operation.function_properties import FunctionProperties
from quam.components import Qubit
from quam.components.macro import QubitMacro
from quam.core import quam_dataclass


def test_operation_initialization():
    def sample_op(qubit: Qubit):
        pass

    op = Operation(sample_op)
    assert op.func == sample_op
    assert isinstance(op.properties, FunctionProperties)
    assert op.properties.name == "sample_op"
    assert op.properties.quantum_component_type == Qubit


def test_operation_get_macro(test_qubit):
    def x_gate(qubit: Qubit):
        pass

    op = Operation(x_gate)
    retrieved_macro = op.get_macro(test_qubit)
    assert retrieved_macro == test_qubit.macros["x_gate"]


def test_operation_get_macro_missing(test_qubit):
    def missing_op(qubit: Qubit):
        pass

    op = Operation(missing_op)
    with pytest.raises(KeyError, match="Operation 'missing_op' is not implemented"):
        op.get_macro(test_qubit)


def test_operation_call(test_qubit):
    def test_op(qubit: Qubit, amplitude: float = 1.0):
        pass

    op = Operation(test_op)
    result = op(test_qubit, amplitude=0.5)

    # Check results
    assert result[0] == test_qubit  # First element should be the qubit
    assert result[1] == ()  # No positional args
    assert result[2] == {"amplitude": 0.5}  # Keyword args


def test_operation_call_invalid_component():
    def test_op(qubit: Qubit):
        pass

    op = Operation(test_op)

    # Try to call with wrong type
    with pytest.raises(ValueError, match="First argument to test_op must be a Qubit"):
        op("not_a_qubit")


def test_operation_call_no_args():
    def test_op(qubit: Qubit):
        pass

    op = Operation(test_op)

    # Try to call with no arguments
    with pytest.raises(
        ValueError, match="Operation test_op requires at least one argument"
    ):
        op()


def test_operation_call_multiple_args(test_qubit):
    def test_op(qubit: Qubit, arg1: float, arg2: str):
        pass

    op = Operation(test_op)
    result = op(test_qubit, 1.0, "test")

    assert result[0] == test_qubit
    assert result[1] == (1.0, "test")
    assert result[2] == {}  # No keyword args


@quam_dataclass
class TestMacro2(QubitMacro):
    """Test macro class that requires a positional argument"""

    def apply(self, required_arg, **kwargs):
        # Return inputs to verify they were passed correctly
        return (self.qubit, (required_arg,), kwargs)


def test_operation_call_out_of_order_kwargs(test_qubit):
    def test_op(qubit: Qubit, arg1: float, arg2: str = "default"):
        pass

    # Use TestMacro2 which requires a positional argument
    macro = TestMacro2()
    test_qubit.macros["test_op"] = macro

    op = Operation(test_op)
    # Pass arg2 before arg1, making arg1 (a positional arg) into a kwarg
    result = op(test_qubit, arg2="test", required_arg=1.0)

    assert result[0] == test_qubit
    assert result[1] == (1.0,)  # arg1 as positional arg
    assert result[2] == {"arg2": "test"}  # arg2 as kwarg


def test_measure_operation(test_qubit):
    from quam.utils.qua_types import QuaVariableBool

    def measure(qubit: Qubit, **kwargs) -> QuaVariableBool:
        pass

    op = Operation(measure)

    assert op.properties.return_type == QuaVariableBool
    assert op.func == measure



import dataclasses
from dataclasses import dataclass, fields, is_dataclass
import functools
import sys
import warnings
from typing import Dict, Union, ClassVar, get_type_hints

__all__ = ["patch_dataclass", "get_dataclass_attr_annotations"]


class REQUIRED:
    """Flag used by `quam_dataclass` when a required dataclass arg needs a kwarg"""

    ...


def get_dataclass_attr_annotations(
    cls_or_obj: Union[type, object],
) -> Dict[str, Dict[str, type]]:
    """Get the attributes and annotations of a dataclass

    Args:
        cls: The dataclass to get the attributes of.

    Returns:
        A dictionary where the keys are "required", "optional" and "allowed".
            - "required": Required attributes of the class.
            - "optional": Optional attributes of the class, i.e. with a default value.
            - "allowed": allowed attributes of the class := "required" + "optional".
        For each key, the values are dictionaries with the attribute names as keys
        and the attribute types as values.
    """

    annotated_attrs = get_type_hints(cls_or_obj)

    annotated_attrs.pop("_root", None)
    annotated_attrs.pop("_references", None)
    annotated_attrs.pop("_skip_attrs", None)
    annotated_attrs.pop("parent", None)
    annotated_attrs.pop("config_settings", None)
    annotated_attrs.pop("_value_annotation", None)
    annotated_attrs.pop("_initialized", None)

    attr_annotations = {"required": {}, "optional": {}}
    for attr, attr_type in annotated_attrs.items():
        if getattr(attr_type, "__origin__", None) == ClassVar:
            continue
        # TODO Try to combine with third elif statement
        if getattr(cls_or_obj, attr, None) is REQUIRED:  # See "patch_dataclass()"
            if attr not in getattr(cls_or_obj, "__dataclass_fields__", {}):
                attr_annotations["required"][attr] = attr_type
            else:
                field_attr = cls_or_obj.__dataclass_fields__[attr]
                if field_attr.default_factory is not dataclasses.MISSING:
                    attr_annotations["optional"][attr] = attr_type
                else:
                    attr_annotations["required"][attr] = attr_type
        elif hasattr(cls_or_obj, attr):
            attr_annotations["optional"][attr] = attr_type
        elif attr in getattr(cls_or_obj, "__dataclass_fields__", {}):
            data_field = cls_or_obj.__dataclass_fields__[attr]
            if data_field.default_factory is not dataclasses.MISSING:
                attr_annotations["optional"][attr] = attr_type
            else:
                attr_annotations["required"][attr] = attr_type
        else:
            attr_annotations["required"][attr] = attr_type
    attr_annotations["allowed"] = {
        **attr_annotations["required"],
        **attr_annotations["optional"],
    }
    return attr_annotations


def dataclass_field_has_default(field: dataclasses.field) -> bool:
    """Check if a dataclass field has a default value"""
    if field.default is not dataclasses.MISSING:
        return True
    elif field.default_factory is not dataclasses.MISSING:
        return True
    return False


def handle_inherited_required_fields(cls):
    """Adds a default REQUIRED flag for dataclass fields when necessary

    see quam_dataclass docs for details
    """
    if not is_dataclass(cls):
        return

    # Check if dataclass has fields with default value
    optional_fields = [
        field.name
        for field in dataclasses.fields(cls)
        if dataclass_field_has_default(field)
    ]
    if not optional_fields:
        # All fields of the dataclass are required, we don't have to handle situations
        # where the parent class has fields with default values and the subclass has
        # required fields.
        return

    # Check if class (not parents) has required fields
    for attr, attr_type in cls.__annotations__.items():
        if attr in cls.__dict__:
            continue
        if attr in optional_fields:
            continue
        if getattr(attr_type, "__origin__", None) is ClassVar:
            continue
        setattr(cls, attr, REQUIRED)


def _quam_patched_dataclass(cls=None, kw_only: bool = False, eq: bool = True):
    """Patch around Python dataclass to ensure that dataclass subclassing works

    Prior to Python 3.10, it was not possible for a dataclass to be a subclass of
    another dataclass when
    - the parent dataclass has an arg with default
    - the child dataclass has a required arg

    From Python 3.10, this was fixed by including the flag @dataclass(kw_only=True).
    To ensure QUAM remains compatible with Python <3.10, we include a method to patch
    the dataclass such that it still works in the case described above.

    We achieve this by first checking if the above condition is met. If so, all the
    args without a default value receive a default REQUIRED flag. The post_init method
    is then overridden such that an error is raised whenever an attribute still has
    the REQUIRED flag after instantiation.

    The python dataclass can be overridden by calling patch_dataclass.
    """
    if cls is None:
        return functools.partial(_quam_patched_dataclass, kw_only=kw_only, eq=eq)

    handle_inherited_required_fields(cls)

    post_init_method = getattr(cls, "__post_init__", None)

    def __post_init__(self, *args, **kwargs):
        with warnings.catch_warnings():  # Ignore warnings of failed references
            warnings.filterwarnings("ignore", category=UserWarning)
            for f in fields(self):
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore",
                        message="^Could not get reference*",
                        category=UserWarning,
                    )
                    from quam.utils.exceptions import InvalidReferenceError

                    try:
                        attr_val = getattr(self, f.name, None)
                    except InvalidReferenceError:
                        attr_val = None
                if attr_val is REQUIRED:
                    raise TypeError(
                        f"Please provide {cls.__name__}.{f.name} as it is a"
                        " required arg"
                    )

        if post_init_method is not None:
            post_init_method(self, *args, **kwargs)

    cls.__post_init__ = __post_init__
    cls_dataclass = dataclass(cls, eq=False)

    return cls_dataclass


def patch_dataclass(module_name):
    """Patch Python dataclass within a file to allow subclasses have args

    Note:
        Patch is only applied when Python < 3.10.

    Note:
        This function should be called at the top of a file, before dataclasses are
        defined:
        ```
        patch_dataclass(__name__)  # Ensure dataclass "kw_only" also works with python < 3.10
        ```

    Prior to Python 3.10, it was not possible for a dataclass to be a subclass of
    another dataclass when
    - the parent dataclass has an arg with default
    - the child dataclass has a required arg

    From Python 3.10, this was fixed by including the flag @dataclass(kw_only=True).
    To ensure QUAM remains compatible with Python <3.10, we include a method to patch
    the dataclass such that it still works in the case described above.

    We achieve this by first checking if the above condition is met. If so, all the
    args without a default value receive a default REQUIRED flag. The post_init method
    is then overridden such that an error is raised whenever an attribute still has
    the REQUIRED flag after instantiation.

    Note:
        The python dataclass is patched in a non-standard way by calling `setattr`
        on the module. This is done to ensure that the patch is not recognized by any
        static analysis tools, such as mypy. This is necessary as mypy otherwise will
        no longer recognize the dataclass as a dataclass.
    """
    DeprecationWarning(
        "patch_dataclass is deprecated and will be removed in QUAM v1.0. "
        "Please use 'from quam.core import quam_dataclass' as a decorator instead of "
        "the regular Python dataclass."
    )
    if sys.version_info.minor < 10:
        setattr(sys.modules[module_name], "dataclass", _quam_patched_dataclass)



from quam.core import QuamComponent, QuamRoot, quam_dataclass


@quam_dataclass
class EmptyRoot(QuamRoot):
    pass


def test_empty_root(qua_config):
    root = EmptyRoot()
    cfg = root.generate_config()

    assert cfg == qua_config


@quam_dataclass
class Component(QuamComponent):
    name: str

    def apply_to_config(self, config: dict) -> None:
        config.setdefault("list", [])
        config["list"].append(self.name)


@quam_dataclass
class Root(QuamRoot):
    first_component: Component
    second_component: Component


def test_generate_basic_config():
    root = Root(
        first_component=Component(name="first"),
        second_component=Component(name="second"),
    )
    config = root.generate_config()
    assert config["list"] == ["first", "second"]


def test_generate_after():
    root = Root(
        first_component=Component(name="first"),
        second_component=Component(name="second"),
    )

    root.first_component.config_settings = {"after": ["#/second_component"]}

    config = root.generate_config()
    assert config["list"] == ["second", "first"]


def test_generate_before():
    root = Root(
        first_component=Component(name="first"),
        second_component=Component(name="second"),
    )

    root.second_component.config_settings = {"before": ["#/first_component"]}

    config = root.generate_config()
    assert config["list"] == ["second", "first"]


def test_generate_after_property():
    @quam_dataclass
    class ComponentProperty(QuamComponent):
        name: str

        def apply_to_config(self, config: dict) -> None:
            config.setdefault("list", [])
            config["list"].append(self.name)

        @property
        def config_settings(self):
            return {"after": [self.get_root().second_component]}

    root = Root(
        first_component=ComponentProperty(name="first"),
        second_component=Component(name="second"),
    )

    config = root.generate_config()
    assert config["list"] == ["second", "first"]



from abc import ABC
from typing import Any, ClassVar, Dict, Optional

from quam.components.ports.base_ports import BasePort, FEMPort, OPXPlusPort
from quam.core import quam_dataclass


__all__ = [
    "LFAnalogInputPort",
    "OPXPlusAnalogInputPort",
    "LFFEMAnalogInputPort",
    "MWFEMAnalogInputPort",
]


@quam_dataclass
class LFAnalogInputPort(BasePort, ABC):
    fem_type: ClassVar[str] = "LF"
    port_type: ClassVar[str] = "analog_input"

    offset: Optional[float] = None
    gain_db: int = 0
    shareable: bool = False

    def get_port_properties(self):
        port_cfg = {
            "gain_db": self.gain_db,
            "shareable": self.shareable,
        }
        if self.offset is not None:
            port_cfg["offset"] = self.offset
        return port_cfg


@quam_dataclass
class OPXPlusAnalogInputPort(LFAnalogInputPort, OPXPlusPort):
    pass


@quam_dataclass
class LFFEMAnalogInputPort(LFAnalogInputPort, FEMPort):
    sampling_rate: float = 1e9  # Either 1e9 or 2e9

    def get_port_properties(self) -> Dict[str, Any]:
        port_properties = super().get_port_properties()
        port_properties["sampling_rate"] = self.sampling_rate
        return port_properties


@quam_dataclass
class MWFEMAnalogInputPort(FEMPort):
    fem_type: ClassVar[str] = "MW"
    port_type: ClassVar[str] = "analog_input"

    band: int
    downconverter_frequency: float
    gain_db: Optional[int] = None
    sampling_rate: float = 1e9  # Either 1e9 or 2e9
    shareable: bool = False

    def get_port_properties(self) -> Dict[str, Any]:
        port_properties = {
            "band": self.band,
            "downconverter_frequency": self.downconverter_frequency,
            "sampling_rate": self.sampling_rate,
            "shareable": self.shareable,
        }
        if self.gain_db is not None:
            port_properties["gain_db"] = self.gain_db
        return port_properties



from abc import ABC
from typing import Any, ClassVar, Dict, Literal

from quam.components.ports.base_ports import BasePort, FEMPort, OPXPlusPort
from quam.core import quam_dataclass


__all__ = ["DigitalOutputPort", "OPXPlusDigitalOutputPort", "FEMDigitalOutputPort"]


@quam_dataclass
class DigitalOutputPort(BasePort, ABC):
    port_type: ClassVar[str] = "digital_output"

    inverted: bool = False
    shareable: bool = False

    def get_port_properties(self) -> Dict[str, Any]:
        return {"inverted": self.inverted, "shareable": self.shareable}


@quam_dataclass
class OPXPlusDigitalOutputPort(DigitalOutputPort, OPXPlusPort):
    pass


@quam_dataclass
class FEMDigitalOutputPort(DigitalOutputPort, FEMPort):
    level: Literal["TTL", "LVTTL"] = "LVTTL"

    def get_port_properties(self) -> Dict[str, Any]:
        port_properties = super().get_port_properties()
        port_properties["level"] = self.level
        return port_properties



from collections import UserList
import pytest

from quam.core import *


def test_quam_list_inheritance():
    quam_list = QuamList()
    assert isinstance(quam_list, QuamBase)
    assert not isinstance(quam_list, list)
    assert isinstance(quam_list, QuamList)
    assert isinstance(quam_list, UserList)


def test_quam_list_equality():
    assert QuamList() == QuamList()
    assert QuamList() != QuamList([1, 2, 3])
    assert QuamList([1, 2, 3]) == QuamList([1, 2, 3])
    assert QuamList([1, 2, 3]) != QuamList([1, 2, 3, 4])

    assert QuamList() == []
    assert QuamList() != [1, 2, 3]
    assert QuamList([1, 2, 3]) == [1, 2, 3]
    assert QuamList([1, 2, 3]) != [1, 2, 3, 4]


def test_quam_list_properties():
    quam_list = QuamList()
    assert quam_list.data == []
    assert quam_list == []

    quam_list += [1, 2, 3]
    assert list(quam_list) == [1, 2, 3]
    assert quam_list == [1, 2, 3]

    assert quam_list.data == [1, 2, 3]

    quam_list2 = QuamList([4, 5, 6])
    assert list(quam_list2) == [4, 5, 6]
    assert quam_list2.data == [4, 5, 6]


def test_quam_nested_list_instantiation_explicit():
    quam_list = QuamList([1, 2, QuamList([3, 4, 5])])
    assert list(quam_list) == [1, 2, [3, 4, 5]]
    assert quam_list.data == [1, 2, [3, 4, 5]]
    assert quam_list[2] == [3, 4, 5]
    assert isinstance(quam_list[2], QuamList)


def test_quam_nested_list_instantiation_implicit():
    quam_list = QuamList([1, 2, [3, 4, 5]])
    assert list(quam_list) == [1, 2, [3, 4, 5]]
    assert quam_list.data == [1, 2, [3, 4, 5]]
    assert quam_list[2] == [3, 4, 5]
    assert isinstance(quam_list[2], QuamList)


def test_list_parent(BareQuamComponent):
    quam_list = QuamList([BareQuamComponent()])
    assert quam_list[0].parent is quam_list

    quam_list.append(BareQuamComponent())
    assert quam_list[1].parent is quam_list

    quam_list.extend([BareQuamComponent(), BareQuamComponent()])
    assert quam_list[2].parent is quam_list

    quam_list[0] = BareQuamComponent()
    assert quam_list[0].parent is quam_list


def test_list_get_attrs_error():
    quam_list = QuamList()
    with pytest.raises(NotImplementedError):
        quam_list.get_attrs()


def test_list_to_dict():
    quam_list = QuamList([1, 2, 3])
    assert quam_list.to_dict() == [1, 2, 3]


qua_config_template = {
    "controllers": {},
    "elements": {},
    "pulses": {
        "const_pulse": {
            "operation": "control",
            "length": 1000,
            "waveforms": {"I": "const_wf", "Q": "zero_wf"},
        }
    },
    "waveforms": {
        "zero_wf": {"type": "constant", "sample": 0.0},
        "const_wf": {"type": "constant", "sample": 0.1},
    },
    "digital_waveforms": {
        "ON": {"samples": [[1, 0]]}
    },  # TODO Technically this waveform isn't used
    "integration_weights": {},
    "mixers": {},
    "oscillators": {},
}



import pytest
import json
from copy import deepcopy

from quam.components import *
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort
from quam.core import quam_dataclass, QuamRoot


@pytest.fixture
def bare_cfg():
    return {
        "controllers": {},
        "elements": {},
    }


def test_single_channel_no_name(bare_cfg):
    channel = SingleChannel(opx_output=("con1", 1))

    with pytest.raises(AttributeError):
        channel.apply_to_config(bare_cfg)

    channel.id = "channel"
    channel.apply_to_config(bare_cfg)


def test_single_channel_offset(bare_cfg):
    channel = SingleChannel(id="channel", opx_output=("con1", 1))

    cfg = deepcopy(bare_cfg)
    channel.apply_to_config(cfg)

    expected_cfg = {
        "controllers": {
            "con1": {
                "analog_outputs": {1: {"delay": 0, "shareable": False}},
            }
        },
        "elements": {
            "channel": {
                "singleInput": {"port": ("con1", 1)},
                "operations": {},
            }
        },
    }
    assert cfg == expected_cfg

    cfg = deepcopy(bare_cfg)
    channel.opx_output_offset = 0.1
    expected_cfg["controllers"]["con1"]["analog_outputs"][1]["offset"] = 0.1
    channel.apply_to_config(cfg)
    assert cfg == expected_cfg


def test_single_channel_differing_offsets(bare_cfg):
    channel1 = SingleChannel(id="channel", opx_output=("con1", 1))
    channel2 = SingleChannel(id="channel2", opx_output=("con1", 1))

    cfg = deepcopy(bare_cfg)
    channel1.apply_to_config(cfg)
    channel2.apply_to_config(cfg)
    assert cfg["controllers"]["con1"]["analog_outputs"][1] == {
        "delay": 0,
        "shareable": False,
    }

    channel1.opx_output_offset = 0.1

    cfg = deepcopy(bare_cfg)
    channel1.apply_to_config(cfg)
    channel2.apply_to_config(cfg)
    assert cfg["controllers"]["con1"]["analog_outputs"][1]["offset"] == 0.1

    channel2.opx_output_offset = 0.1

    cfg = deepcopy(bare_cfg)
    channel1.apply_to_config(cfg)
    channel2.apply_to_config(cfg)
    assert cfg["controllers"]["con1"]["analog_outputs"][1]["offset"] == 0.1

    channel2.opx_output_offset = 0.2

    cfg = deepcopy(bare_cfg)
    channel1.apply_to_config(cfg)
    with pytest.warns(UserWarning):
        channel2.apply_to_config(cfg)

    cfg = deepcopy(bare_cfg)
    channel1.apply_to_config(cfg)
    channel2.opx_output_offset = 0.1 + 0.5e-4
    channel2.apply_to_config(cfg)
    assert cfg["controllers"]["con1"]["analog_outputs"][1]["offset"] == 0.1 + 0.5e-4


def test_single_channel_offset_quam(qua_config):
    @quam_dataclass
    class QuamTest(QuamRoot):
        channel: SingleChannel

    channel = SingleChannel(id="channel", opx_output=("con1", 1))
    machine = QuamTest(channel=channel)

    cfg = machine.generate_config()

    qua_config["controllers"] = {
        "con1": {
            "analog_outputs": {1: {"delay": 0, "shareable": False, "offset": 0.0}},
        }
    }
    qua_config["elements"] = {
        "channel": {
            "singleInput": {"port": ("con1", 1)},
            "operations": {},
        }
    }

    assert cfg == qua_config

    channel.opx_output_offset = 0.1

    cfg = machine.generate_config()
    qua_config["controllers"]["con1"]["analog_outputs"][1]["offset"] = 0.1
    assert cfg == qua_config


def test_single_channel_set_dc_offset(mocker):
    mocker.patch("quam.components.channels.set_dc_offset")

    channel = SingleChannel(id="channel", opx_output=("con1", 1))
    channel.set_dc_offset(0.5)

    from quam.components.channels import set_dc_offset

    set_dc_offset.assert_called_once_with(
        element="channel", element_input="single", offset=0.5
    )


def test_instantiate_single_channel():
    from quam.core.quam_instantiation import instantiate_quam_class

    channel = SingleChannel(opx_output=("con1", 1))
    d = channel.to_dict()

    d_json = json.dumps(d)

    d_loaded = json.loads(d_json)

    instantiate_quam_class(SingleChannel, d_loaded)


def test_generate_config(qua_config):
    channel = SingleChannel(
        id="out_channel",
        opx_output=("con1", 1),
    )

    channel.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_outputs": {
                1: {"delay": 0, "shareable": False},
            },
        }
    }

    assert qua_config["elements"] == {
        "out_channel": {"operations": {}, "singleInput": {"port": ("con1", 1)}}
    }


def test_generate_config_port(qua_config):
    channel = SingleChannel(
        id="out_channel",
        opx_output=OPXPlusAnalogOutputPort("con1", 1),
    )

    channel.apply_to_config(qua_config)
    channel.opx_output.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_outputs": {
                1: {"delay": 0, "shareable": False},
            },
        }
    }

    assert qua_config["elements"] == {
        "out_channel": {"operations": {}, "singleInput": {"port": ("con1", 1)}}
    }



from typing import Optional
from quam.core import *
from quam.utils.string_reference import *


@quam_dataclass
class BareQuamRoot(QuamRoot):
    a: int = 4
    quam_list: Optional[QuamList] = None


def test_referencing_from_list():
    quam_root = BareQuamRoot()
    quam_root.quam_list = []
    assert isinstance(quam_root.quam_list, QuamList)

    quam_root.a = 42
    quam_root.quam_list.append("#/a")
    assert quam_root.quam_list[0] == 42
    assert quam_root.quam_list[:] == [42]

    quam_root.quam_list.append("#/b")
    assert quam_root.quam_list[1] == "#/b"
    quam_root.b = 43
    assert quam_root.quam_list[1] == 43

    assert list(quam_root.quam_list) == [42, 43]


def test_referencing_to_list():
    quam_root = BareQuamRoot()
    quam_root.quam_list = [42, 43]

    assert quam_root._get_referenced_value("#/quam_list/1") == 43
    assert quam_root.quam_list._get_referenced_value("#/quam_list/0") == 42
    assert quam_root._get_referenced_value("#/quam_list/0") == 42
    assert quam_root.quam_list._get_referenced_value("#./1") == 43

    quam_root.quam_list[0] = 44
    assert quam_root.quam_list[0] == 44
    assert quam_root._get_referenced_value("#/quam_list/0") == 44
    assert quam_root.quam_list._get_referenced_value("#/quam_list/0") == 44
    assert quam_root.quam_list._get_referenced_value("#./0") == 44



from typing import Dict, List
import pytest
from quam.core.quam_classes import *
from quam.core.quam_classes import _get_value_annotation


def test_value_annotation_nonexisting():
    @quam_dataclass
    class TestQuam(QuamComponent):
        int_val: int
        str_val: str

    test_quam = TestQuam(int_val=1, str_val="test")
    assert _get_value_annotation(test_quam, "nonexisting") is None
    assert _get_value_annotation(test_quam, "int_val") is None
    assert _get_value_annotation(test_quam, "str_val") is None


def test_value_annotation_dict():
    @quam_dataclass
    class TestQuam(QuamComponent):
        int_val: int
        str_val: str
        d1: dict
        d2: Dict[str, int]

    test_quam = TestQuam(int_val=1, str_val="test", d1={"a": 1}, d2={"a": 1})
    assert _get_value_annotation(test_quam, "d1") is None
    assert _get_value_annotation(test_quam, "str_val") is None
    assert _get_value_annotation(test_quam, "d2") == int


def test_value_annotation_list():
    @quam_dataclass
    class TestQuam(QuamComponent):
        int_val: int
        str_val: str
        l1: list
        l2: List[int]

    test_quam = TestQuam(int_val=1, str_val="test", l1=[1, 2, 3], l2=[1, 2, 3])
    assert _get_value_annotation(test_quam, "l1") is None
    assert _get_value_annotation(test_quam, "str_val") is None
    assert _get_value_annotation(test_quam, "l2") == int


def test_value_annotation_bare_quam_component():
    @quam_dataclass
    class TestQuam(QuamComponent): ...

    test_quam = TestQuam()

    assert _get_value_annotation(test_quam, "attr") is None
    assert _get_value_annotation(TestQuam, "attr") is None


def test_value_annotation_bare_quam_root():
    from quam.core.quam_classes import _get_value_annotation

    @quam_dataclass
    class TestQuam(QuamRoot): ...

    assert _get_value_annotation(TestQuam, "attr") is None

    assert _get_value_annotation(TestQuam(), "attr") is None


def test_type_hints_empty_dataclass():
    @quam_dataclass
    class C: ...

    assert _get_value_annotation(C, "attr") is None

    assert _get_value_annotation(C(), "attr") is None



import pytest
from quam.components.ports.analog_inputs import OPXPlusAnalogInputPort
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort


def test_opx_plus_analog_output_port():
    with pytest.raises(TypeError):
        OPXPlusAnalogOutputPort()

    port = OPXPlusAnalogOutputPort("con1", 2)
    assert port.controller_id == "con1"
    assert port.port_id == 2
    assert port.port_tuple == ("con1", 2)
    assert port.port_type == "analog_output"
    assert port.offset is None
    assert port.delay == 0
    assert port.crosstalk is None
    assert port.feedforward_filter is None
    assert port.feedback_filter is None
    assert port.shareable == False

    assert port.get_port_properties() == {
        "delay": 0,
        "shareable": False,
    }

    assert port.get_port_properties() == {
        "delay": 0,
        "shareable": False,
    }

    cfg = {"controllers": {}}
    port.apply_to_config(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "analog_outputs": {
                    2: {
                        "delay": 0,
                        "shareable": False,
                    }
                }
            }
        }
    }


def test_opx_plus_analog_input_port():
    with pytest.raises(TypeError):
        OPXPlusAnalogInputPort()

    port = OPXPlusAnalogInputPort("con1", 2)
    assert port.controller_id == "con1"
    assert port.port_id == 2
    assert port.port_tuple == ("con1", 2)
    assert port.port_type == "analog_input"
    assert port.offset is None
    assert port.gain_db == 0
    assert port.shareable == False

    assert port.get_port_properties() == {
        "gain_db": 0,
        "shareable": False,
    }

    assert port.get_port_properties() == {
        "gain_db": 0,
        "shareable": False,
    }

    cfg = {"controllers": {}}
    port.apply_to_config(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "analog_inputs": {
                    2: {
                        "gain_db": 0,
                        "shareable": False,
                    }
                }
            }
        }
    }



from typing import Any, Callable, TypeVar
import functools

from quam.core.macro.base_macro import BaseMacro


__all__ = ["method_macro", "MethodMacro"]


T = TypeVar("T", bound=Callable)


class MethodMacro(BaseMacro):
    """Decorator that marks methods which should be exposed as macros."""

    def __init__(self, func: T) -> None:
        functools.wraps(func)(self)
        self.func = func
        self.instance = None

    def __get__(self, instance, owner):
        # Store the instance to which this method is bound
        self.instance = instance
        return self

    def apply(self, *args, **kwargs) -> Any:
        """Implements BaseMacro.apply by calling the wrapped function"""
        if self.instance is not None:
            # Call the function with the instance as the first argument
            return self.func(self.instance, *args, **kwargs)
        return self.func(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if args and args[0] is self.instance:
            args = args[1:]
        return self.apply(*args, **kwargs)

    @staticmethod
    def is_macro_method(obj: Any) -> bool:
        return isinstance(obj, MethodMacro)


# Lower-case alias for MethodMacro for more pythonic decorator
method_macro = MethodMacro



from quam.components.channels import InSingleOutIQChannel
from quam.components.ports.analog_inputs import OPXPlusAnalogInputPort
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort
from quam.utils.dataclass import get_dataclass_attr_annotations


def test_in_single_channel_attr_annotations():
    attr_annotations = get_dataclass_attr_annotations(InSingleOutIQChannel)
    assert set(attr_annotations["required"]) == {
        "opx_output_I",
        "opx_output_Q",
        "opx_input",
        "frequency_converter_up",
    }
    assert set(attr_annotations["optional"]) == {
        "operations",
        "sticky",
        "intermediate_frequency",
        "LO_frequency",
        "RF_frequency",
        "opx_output_offset_I",
        "opx_output_offset_Q",
        "id",
        "digital_outputs",
        "opx_input_offset",
        "time_of_flight",
        "smearing",
        "thread",
        "core",
        "time_tagging",
    }


def test_generate_config(qua_config):
    channel = InSingleOutIQChannel(
        id="in_out_channel",
        opx_input=("con1", 1),
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
    )

    channel.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_inputs": {
                1: {"gain_db": 0, "shareable": False},
            },
            "analog_outputs": {
                1: {"delay": 0, "shareable": False},
                2: {"delay": 0, "shareable": False},
            },
        }
    }

    assert qua_config["elements"] == {
        "in_out_channel": {
            "intermediate_frequency": 0.0,
            "mixInputs": {
                "I": ("con1", 1),
                "Q": ("con1", 2),
            },
            "operations": {},
            "outputs": {"out1": ("con1", 1)},
            "smearing": 0,
            "time_of_flight": 140,
        }
    }


def test_generate_config(qua_config):
    channel = InSingleOutIQChannel(
        id="in_out_channel",
        opx_input=OPXPlusAnalogInputPort("con1", 1),
        opx_output_I=OPXPlusAnalogOutputPort("con1", 1),
        opx_output_Q=OPXPlusAnalogOutputPort("con1", 2),
        frequency_converter_up=None,
    )

    channel.apply_to_config(qua_config)
    channel.opx_input.apply_to_config(qua_config)
    channel.opx_output_I.apply_to_config(qua_config)
    channel.opx_output_Q.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_inputs": {
                1: {"gain_db": 0, "shareable": False},
            },
            "analog_outputs": {
                1: {"delay": 0, "shareable": False},
                2: {"delay": 0, "shareable": False},
            },
        }
    }

    assert qua_config["elements"] == {
        "in_out_channel": {
            "mixInputs": {
                "I": ("con1", 1),
                "Q": ("con1", 2),
            },
            "operations": {},
            "outputs": {"out1": ("con1", 1)},
            "smearing": 0,
            "time_of_flight": 140,
        }
    }



"""Comprehensive unit test suite for reference chain functionality.

Tests reference chain following through multiple levels, QuamList/QuamDict
with reference chains, set_at_reference() with chains, and edge cases.
"""

import pytest

from quam.core.quam_classes import QuamBase, QuamRoot, QuamList, QuamDict, quam_dataclass


# Fixture classes
@quam_dataclass
class SimpleTarget(QuamBase):
    value: int = 0


@quam_dataclass
class ChainableValue(QuamBase):
    target: "ChainableValue" = None
    alias_ref: str = ""


@quam_dataclass
class ListContainer(QuamBase):
    values: QuamList = None
    list_ref: str = ""


@quam_dataclass
class DictContainer(QuamBase):
    data: QuamDict = None
    dict_ref: str = ""


@quam_dataclass
class QuamRootForTesting(QuamRoot):
    target: SimpleTarget = None
    chainable: ChainableValue = None
    list_container: ListContainer = None
    dict_container: DictContainer = None
    simple_value: int = 0


# Reference chain following tests
def test_simple_reference_no_chain():
    """Direct reference to a simple value - no chain to follow."""
    root = QuamRootForTesting(target=SimpleTarget(value=42))
    assert root.target.value == 42


def test_single_reference_chain_one_hop():
    """Reference chain with one hop: attr -> target."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=42),
        chainable=ChainableValue(),
    )
    root.chainable.alias_ref = "#/target/value"

    assert root.chainable.alias_ref == 42


def test_double_reference_chain_two_hops():
    """Reference chain with two hops: attr1 -> attr2 (which is ref) -> target."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=123),
        chainable=ChainableValue(target=ChainableValue(alias_ref="#/target/value")),
    )
    root.chainable.alias_ref = "#/chainable/target/alias_ref"

    assert root.chainable.alias_ref == 123


def test_triple_reference_chain():
    """Reference chain with three hops."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=999),
        chainable=ChainableValue(
            target=ChainableValue(target=ChainableValue(alias_ref="#/target/value"))
        ),
    )
    root.chainable.target.alias_ref = "#/chainable/target/target/alias_ref"

    assert root.chainable.target.alias_ref == 999


def test_empty_reference_not_a_chain():
    """Empty string is not a reference - no chain to follow."""
    root = QuamRootForTesting(simple_value=42)
    assert root.simple_value == 42


# QuamList with reference tests
def test_list_element_simple_value_no_chain():
    """List element contains a simple value - no chain."""
    container = ListContainer(values=QuamList())
    container.values.extend([42, 43, 44])

    assert container.values[0] == 42
    assert container.values[1] == 43


def test_list_element_is_reference_to_value():
    """List element is a reference string to a simple value."""
    root = QuamRootForTesting(
        simple_value=555,
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.extend(["#/simple_value"])

    assert root.list_container.values[0] == 555


def test_list_element_is_reference_to_another_list_index():
    """List element references another list index."""
    root = QuamRootForTesting(
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.extend(["#/list_container/values/1", 999])

    assert root.list_container.values[0] == 999
    assert root.list_container.values[1] == 999


def test_list_element_chain_list_ref_to_alias_to_target():
    """List element chains through multiple hops."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=888),
        chainable=ChainableValue(alias_ref="#/target/value"),
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.extend(["#/chainable/alias_ref"])

    assert root.list_container.values[0] == 888


def test_list_element_with_broken_reference_chain():
    """Broken reference in list chain should fail gracefully."""
    root = QuamRootForTesting(
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.extend(["#/nonexistent/value"])

    result = root.list_container.values[0]
    assert result == "#/nonexistent/value" or isinstance(result, str)


def test_list_element_deep_chain():
    """Deep reference chain through multiple levels in list."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=666),
        chainable=ChainableValue(target=ChainableValue(alias_ref="#/target/value")),
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.append("#/chainable/target/alias_ref")

    assert root.list_container.values[0] == 666


def test_list_with_mixed_references_and_values():
    """List with both reference elements and regular values."""
    root = QuamRootForTesting(
        simple_value=111,
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.extend(
        [
            "#/simple_value",
            222,
            "#/simple_value",
        ]
    )

    assert root.list_container.values[0] == 111
    assert root.list_container.values[1] == 222
    assert root.list_container.values[2] == 111


def test_list_element_relative_reference_chain():
    """Relative reference in list chain."""
    root = QuamRootForTesting(
        simple_value=333,
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.append("#/simple_value")

    assert root.list_container.values[0] == 333


def test_list_to_dict_chain():
    """Reference chain from list element to dict value."""
    root = QuamRootForTesting(
        simple_value=444,
        dict_container=DictContainer(data=QuamDict()),
        list_container=ListContainer(values=QuamList()),
    )
    root.dict_container.data["target"] = "#/simple_value"
    root.list_container.values.append("#/dict_container/data/target")

    assert root.list_container.values[0] == 444


# QuamDict with reference tests
def test_dict_value_simple_no_chain():
    """Dict value is simple - no chain to follow."""
    container = DictContainer(data=QuamDict())
    container.data["key1"] = 42

    assert container.data["key1"] == 42


def test_dict_value_is_reference_to_value():
    """Dict value is a reference string to another value."""
    root = QuamRootForTesting(
        simple_value=444,
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["ref_key"] = "#/simple_value"

    assert root.dict_container.data["ref_key"] == 444


def test_dict_value_is_reference_to_another_dict_entry():
    """Dict value references another dict entry."""
    root = QuamRootForTesting(
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["ref_key"] = "#/dict_container/data/target_key"
    root.dict_container.data["target_key"] = 555

    assert root.dict_container.data["ref_key"] == 555


def test_dict_value_chain_dict_ref_to_alias_to_target():
    """Dict value chains through multiple levels."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=666),
        chainable=ChainableValue(alias_ref="#/target/value"),
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["chained_ref"] = "#/chainable/alias_ref"

    assert root.dict_container.data["chained_ref"] == 666


def test_dict_value_broken_reference_chain():
    """Broken reference in dict chain."""
    root = QuamRootForTesting(
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["broken"] = "#/nonexistent/key"

    result = root.dict_container.data["broken"]
    assert result == "#/nonexistent/key" or isinstance(result, str)


def test_dict_value_deep_chain():
    """Deep reference chain in dict values."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=777),
        chainable=ChainableValue(target=ChainableValue(alias_ref="#/target/value")),
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["deep"] = "#/chainable/target/alias_ref"

    assert root.dict_container.data["deep"] == 777


def test_dict_with_mixed_references_and_values():
    """Dict with both reference and regular values."""
    root = QuamRootForTesting(
        simple_value=888,
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["ref"] = "#/simple_value"
    root.dict_container.data["plain"] = 999

    assert root.dict_container.data["ref"] == 888
    assert root.dict_container.data["plain"] == 999


def test_dict_key_lookup_unchanged():
    """Dict key lookup should remain unchanged."""
    root = QuamRootForTesting(
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["key1"] = 42

    assert "key1" in root.dict_container.data
    assert list(root.dict_container.data.keys()) == ["key1"]


def test_dict_attribute_access_with_reference():
    """Attribute access on dict should also follow references."""
    root = QuamRootForTesting(
        simple_value=555,
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data.ref_attr = "#/simple_value"

    assert root.dict_container.data["ref_attr"] == 555
    assert root.dict_container.data.ref_attr == 555


def test_dict_to_list_chain():
    """Reference chain from dict value to list element."""
    root = QuamRootForTesting(
        simple_value=333,
        list_container=ListContainer(values=QuamList()),
        dict_container=DictContainer(data=QuamDict()),
    )
    root.list_container.values.append("#/simple_value")
    root.dict_container.data["list_ref"] = "#/list_container/values/0"

    assert root.dict_container.data["list_ref"] == 333


# set_at_reference with chain tests
def test_set_at_double_reference_chain():
    """set_at_reference through a double reference chain."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=0),
        chainable=ChainableValue(target=ChainableValue(alias_ref="#/target/value")),
    )
    root.chainable.alias_ref = "#/chainable/target/alias_ref"

    root.chainable.set_at_reference("alias_ref", 123)

    assert root.target.value == 123


def test_set_at_triple_reference_chain():
    """set_at_reference through a triple reference chain."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=0),
        chainable=ChainableValue(
            target=ChainableValue(target=ChainableValue(alias_ref="#/target/value"))
        ),
    )
    root.chainable.target.alias_ref = "#/chainable/target/target/alias_ref"

    root.chainable.target.set_at_reference("alias_ref", 789)

    assert root.target.value == 789


def test_set_at_reference_with_list_index():
    """set_at_reference where reference ends with a list index."""
    container = ListContainer(values=QuamList())
    container.values.extend([1, 2, 3])
    container.list_ref = "#./values/0"

    container.set_at_reference("list_ref", 999)

    assert container.values[0] == 999


def test_set_at_reference_with_multiple_list_indices():
    """set_at_reference with different list indices."""
    container = ListContainer(values=QuamList())
    container.values.extend([1, 2, 3])

    container.list_ref = "#./values/0"
    container.list_ref_1 = "#./values/1"

    container.set_at_reference("list_ref", 999)
    container.set_at_reference("list_ref_1", 888)

    assert container.values[0] == 999
    assert container.values[1] == 888
    assert container.values[2] == 3


def test_set_at_reference_nested_list_index():
    """set_at_reference through nested list indices."""

    @quam_dataclass
    class NestedContainerQuam(QuamBase):
        inner_list: QuamList = None

    @quam_dataclass
    class OuterContainerQuam(QuamBase):
        nested: NestedContainerQuam = None
        nested_ref: str = "#./nested/inner_list/0"

    nested = NestedContainerQuam(inner_list=QuamList())
    nested.inner_list.extend([100, 200, 300])

    outer = OuterContainerQuam(nested=nested)

    outer.set_at_reference("nested_ref", 555)

    assert outer.nested.inner_list[0] == 555
    assert outer.nested.inner_list[1] == 200


def test_set_at_absolute_reference():
    """set_at_reference with absolute references."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=0),
    )
    root.abs_ref = "#/target/value"

    root.set_at_reference("abs_ref", 456)

    assert root.target.value == 456
    assert root.get_raw_value("abs_ref") == "#/target/value"


def test_set_at_reference_non_reference_raises():
    """set_at_reference on non-reference now works with new default."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=0),
    )
    root.normal_value = 42

    # With new default (allow_non_reference=True), this should succeed
    root.set_at_reference("normal_value", 123)
    assert root.normal_value == 123

    # Explicitly passing allow_non_reference=False should raise ValueError
    with pytest.raises(ValueError, match="is not a reference"):
        root.set_at_reference("normal_value", 456, allow_non_reference=False)


# Edge cases and error handling
def test_circular_reference_same_object():
    """Circular reference pointing to itself."""
    root = QuamRootForTesting(
        chainable=ChainableValue(),
    )
    root.chainable.alias_ref = "#/chainable/alias_ref"

    result = root.chainable.alias_ref
    assert result == "#/chainable/alias_ref"


def test_missing_intermediate_attribute_in_chain():
    """Missing attribute in middle of chain."""
    root = QuamRootForTesting(
        chainable=ChainableValue(),
    )
    root.chainable.alias_ref = "#/nonexistent/value"

    result = root.chainable.alias_ref
    assert isinstance(result, str)


def test_reference_to_none_value():
    """Reference pointing to a None value."""
    root = QuamRootForTesting(
        target=None,
        chainable=ChainableValue(),
    )
    root.chainable.alias_ref = "#/target"

    assert root.chainable.alias_ref is None


def test_empty_string_reference():
    """Empty string should not be treated as reference."""
    root = QuamRootForTesting(
        chainable=ChainableValue(),
    )
    root.chainable.alias_ref = ""

    assert root.chainable.alias_ref == ""


def test_reference_to_list_entire():
    """Reference pointing to entire list (not indexed)."""
    root = QuamRootForTesting(
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.extend([1, 2, 3])

    root.chainable = ChainableValue()
    root.chainable.alias_ref = "#/list_container/values"

    assert isinstance(root.chainable.alias_ref, (list, QuamList))
    assert list(root.chainable.alias_ref) == [1, 2, 3]


def test_reference_chain_through_dict():
    """Chain going through dict intermediate values."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=111),
        dict_container=DictContainer(data=QuamDict()),
    )
    root.dict_container.data["intermediate"] = "#/target/value"

    root.chainable = ChainableValue()
    root.chainable.alias_ref = "#/dict_container/data/intermediate"

    assert root.chainable.alias_ref == 111


def test_reference_with_list_and_dict_combination():
    """Complex chain through both lists and dicts."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=222),
        dict_container=DictContainer(data=QuamDict()),
        list_container=ListContainer(values=QuamList()),
    )

    root.list_container.values.append("#/target/value")
    root.dict_container.data["elem"] = "#/list_container/values/0"

    root.chainable = ChainableValue()
    root.chainable.alias_ref = "#/dict_container/data/elem"

    assert root.chainable.alias_ref == 222


def test_invalid_reference_raises_error():
    """Invalid reference should raise AttributeError."""
    root = QuamRootForTesting(
        chainable=ChainableValue(),
    )
    root.chainable.alias_ref = "#/nonexistent/path/to/nowhere"

    result = root.chainable.alias_ref
    assert isinstance(result, str)


# Backward compatibility tests
def test_existing_set_at_reference_still_works():
    """Existing set_at_reference behavior unchanged."""

    @quam_dataclass
    class Parent(QuamBase):
        child: SimpleTarget = None
        ref_value: str = "#./child/value"

    parent = Parent(child=SimpleTarget(value=0))
    parent.set_at_reference("ref_value", 42)

    assert parent.child.value == 42
    assert parent.get_raw_value("ref_value") == "#./child/value"


def test_existing_list_access_works():
    """Existing list access without references still works."""
    container = ListContainer(values=QuamList())
    container.values.extend([1, 2, 3])

    assert container.values[0] == 1
    assert container.values[1] == 2
    assert container.values[2] == 3


def test_existing_dict_access_works():
    """Existing dict access without references still works."""
    container = DictContainer(data=QuamDict())
    container.data["key1"] = 42
    container.data["key2"] = 43

    assert container.data["key1"] == 42
    assert container.data["key2"] == 43


def test_raw_value_retrieval_unchanged():
    """get_raw_value() always returns original string."""
    root = QuamRootForTesting(
        simple_value=42,
        chainable=ChainableValue(alias_ref="#/simple_value"),
    )

    assert root.chainable.get_raw_value("alias_ref") == "#/simple_value"
    assert root.chainable.alias_ref == 42


# Integration tests
def test_complex_hierarchy_with_chains():
    """Complex object hierarchy with reference chains at multiple levels."""
    root = QuamRootForTesting(
        simple_value=100,
        target=SimpleTarget(value=200),
        chainable=ChainableValue(
            target=SimpleTarget(value=0), alias_ref="#/target/value"
        ),
        list_container=ListContainer(values=QuamList()),
        dict_container=DictContainer(data=QuamDict()),
    )

    root.chainable.target.value = 300
    root.list_container.values.extend(
        [
            "#/chainable/alias_ref",
            "#/target/value",
        ]
    )
    root.dict_container.data["ref1"] = "#/list_container/values/0"
    root.dict_container.data["ref2"] = "#/list_container/values/1"

    assert root.list_container.values[0] == 200
    assert root.dict_container.data["ref1"] == 200
    assert root.dict_container.data["ref2"] == 200


def test_set_and_read_through_chains():
    """Set value through one chain, read through another."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=0),
        chainable=ChainableValue(alias_ref="#/target/value"),
        list_container=ListContainer(values=QuamList()),
    )
    root.list_container.values.append("#/chainable/alias_ref")

    root.chainable.set_at_reference("alias_ref", 555)

    assert root.list_container.values[0] == 555
    assert root.target.value == 555


def test_update_preserves_all_reference_strings():
    """Updating through chain preserves all intermediate reference strings."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=0),
        chainable=ChainableValue(target=ChainableValue(alias_ref="#/target/value")),
    )
    root.chainable.alias_ref = "#/chainable/target/alias_ref"

    root.chainable.set_at_reference("alias_ref", 777)

    assert root.get_raw_value("chainable") is not None
    assert root.chainable.get_raw_value("alias_ref") == "#/chainable/target/alias_ref"
    assert root.chainable.target.get_raw_value("alias_ref") == "#/target/value"
    assert root.target.value == 777


def test_multiple_paths_to_same_target():
    """Multiple reference chains pointing to the same target."""
    root = QuamRootForTesting(
        simple_value=100,
        chainable=ChainableValue(alias_ref="#/simple_value"),
        list_container=ListContainer(values=QuamList()),
        dict_container=DictContainer(data=QuamDict()),
    )

    root.list_container.values.append("#/simple_value")
    root.dict_container.data["direct"] = "#/simple_value"
    root.dict_container.data["through_chain"] = "#/chainable/alias_ref"

    assert root.list_container.values[0] == 100
    assert root.dict_container.data["direct"] == 100
    assert root.dict_container.data["through_chain"] == 100


def test_chain_with_partial_references():
    """Chain where only intermediate values are references."""
    root = QuamRootForTesting(
        target=SimpleTarget(value=42),
    )

    root.chainable = ChainableValue(target=root.target, alias_ref="#/target/value")

    assert root.chainable.alias_ref == 42



"""Test v2 to v3 migration for include_defaults_in_serialization field."""

import pytest
from quam.config.cli.migrations.v2_v3 import Migrate


def test_forward_migration_adds_serialization_section():
    """Test that forward migration (v2→v3) adds serialization section."""
    v2_config = {
        "quam": {
            "version": 2,
            "state_path": None,
            "raise_error_missing_reference": False,
        }
    }

    v3_config = Migrate.forward(v2_config)

    assert v3_config["quam"]["version"] == 3
    assert "serialization" in v3_config["quam"]
    assert v3_config["quam"]["serialization"]["include_defaults"] is True
    # Ensure other fields are preserved
    assert v3_config["quam"]["state_path"] is None
    assert v3_config["quam"]["raise_error_missing_reference"] is False


def test_backward_migration_removes_serialization_section():
    """Test that backward migration (v3→v2) removes serialization section."""
    v3_config = {
        "quam": {
            "version": 3,
            "state_path": None,
            "raise_error_missing_reference": False,
            "serialization": {"include_defaults": True},
        }
    }

    v2_config = Migrate.backward(v3_config)

    assert v2_config["quam"]["version"] == 2
    assert "serialization" not in v2_config["quam"]
    # Ensure other fields are preserved
    assert v2_config["quam"]["state_path"] is None
    assert v2_config["quam"]["raise_error_missing_reference"] is False


def test_forward_migration_with_state_path():
    """Test forward migration preserves state_path."""
    from pathlib import Path

    v2_config = {
        "quam": {
            "version": 2,
            "state_path": "/path/to/state.json",
            "raise_error_missing_reference": True,
        }
    }

    v3_config = Migrate.forward(v2_config)

    assert v3_config["quam"]["state_path"] == "/path/to/state.json"
    assert v3_config["quam"]["raise_error_missing_reference"] is True
    assert v3_config["quam"]["serialization"]["include_defaults"] is True


def test_migration_versions_are_correct():
    """Test that migration class has correct version numbers."""
    assert Migrate.from_version == 2
    assert Migrate.to_version == 3



from abc import ABC
from typing import Any, ClassVar, Dict, List, Literal, Optional, Tuple

from quam.components.ports.base_ports import BasePort, FEMPort, OPXPlusPort
from quam.core import quam_dataclass

__all__ = [
    "LFAnalogOutputPort",
    "OPXPlusAnalogOutputPort",
    "LFFEMAnalogOutputPort",
    "MWFEMAnalogOutputPort",
]


@quam_dataclass
class LFAnalogOutputPort(BasePort, ABC):
    fem_type: ClassVar[str] = "LF"
    port_type: ClassVar[str] = "analog_output"

    offset: Optional[float] = None
    delay: int = 0
    crosstalk: Optional[Dict[int, float]] = None
    feedforward_filter: Optional[List[float]] = None
    feedback_filter: Optional[List[float]] = None
    shareable: bool = False

    def get_port_properties(self):
        port_properties = {
            "delay": self.delay,
            "shareable": self.shareable,
        }
        if self.crosstalk is not None:
            port_properties["crosstalk"] = dict(self.crosstalk)
        if self.feedforward_filter is not None:
            port_properties.setdefault("filter", {})["feedforward"] = list(
                self.feedforward_filter
            )
        if self.feedback_filter is not None:
            port_properties.setdefault("filter", {})["feedback"] = list(
                self.feedback_filter
            )
        if self.offset is not None:
            port_properties["offset"] = self.offset
        return port_properties


@quam_dataclass
class OPXPlusAnalogOutputPort(LFAnalogOutputPort, OPXPlusPort):
    pass


@quam_dataclass
class LFFEMAnalogOutputPort(LFAnalogOutputPort, FEMPort):
    fem_type: ClassVar[str] = "LF"
    sampling_rate: float = 1e9  # Either 1e9 or 2e9
    upsampling_mode: Literal["mw", "pulse"] = "mw"
    exponential_filter: Optional[List[Tuple[float, float]]] = None
    # high_pass_filter: Optional[float] = None  # Not yet supported
    output_mode: Literal["direct", "amplified"] = "direct"

    def get_port_properties(self) -> Dict[str, Any]:
        port_properties = super().get_port_properties()
        if self.exponential_filter is not None:
            filter_properties = port_properties.setdefault("filter", {})
            filter_properties["exponential"] = list(self.exponential_filter)
            if "feedback" in filter_properties:
                raise ValueError(
                    "LFFEMAnalogOutputPort: Please only specify 'exponential_filter' "
                    "if QOP >=3.3.0, or 'feedback_filter' if QOP < 3.3.0, not both"
                )

        port_properties["sampling_rate"] = self.sampling_rate
        if self.sampling_rate == 1e9:
            port_properties["upsampling_mode"] = self.upsampling_mode
        port_properties["output_mode"] = self.output_mode
        return port_properties


@quam_dataclass
class MWFEMAnalogOutputPort(FEMPort):
    fem_type: ClassVar[str] = "MW"
    port_type: ClassVar[str] = "analog_output"

    band: int
    upconverter_frequency: Optional[float] = None
    upconverters: Optional[Dict[int, Dict[str, float]]] = None
    delay: int = 0
    shareable: bool = False
    sampling_rate: float = 1e9  # Either 1e9 or 2e9
    full_scale_power_dbm: int = -11

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.upconverter_frequency is None and self.upconverters is None:
            raise ValueError(
                "MWAnalogOutputPort: Either upconverter_frequency or upconverters must "
                "be provided"
            )

    def get_port_properties(self) -> Dict[str, Any]:
        port_cfg = {
            "band": self.band,
            "delay": self.delay,
            "shareable": self.shareable,
            "sampling_rate": self.sampling_rate,
            "full_scale_power_dbm": self.full_scale_power_dbm,
        }
        if self.upconverter_frequency is not None:
            port_cfg["upconverter_frequency"] = self.upconverter_frequency
        if self.upconverters is not None:
            port_cfg["upconverters"] = {
                key: dict(val) for key, val in self.upconverters.items()
            }
        return port_cfg



from copy import deepcopy
from pathlib import Path

import click
from click.exceptions import Exit

from qualibrate_config.cli.migrate import migrate_command
from qualibrate_config.cli.utils.content import (
    get_config_file_content,
    simple_write,
    write_config,
)
from qualibrate_config.validation import (
    InvalidQualibrateConfigVersion,
    qualibrate_version_validator,
)
from qualibrate_config.vars import (
    DEFAULT_CONFIG_FILENAME,
    QUALIBRATE_PATH,
)
from qualibrate_config.cli.utils.from_sources import get_config_by_args_mapping

from quam.config.models import QuamConfig, QuamTopLevelConfig
from quam.config.validators import validate_state_path
from quam.config.vars import QUAM_CONFIG_KEY

__all__ = ["config_command"]


@click.command(name="config")
@click.option(
    "--config-path",
    type=click.Path(
        exists=False,
        path_type=Path,
    ),
    default=QUALIBRATE_PATH / DEFAULT_CONFIG_FILENAME,
    show_default=True,
    help=(
        "Path to the configuration file. If the path points to a file, it will "
        "be read and the old configuration will be reused, except for the "
        "variables specified by the user. If the file does not exist, a new one"
        " will be created. If the path points to a directory, a check will be "
        "made to see if files with the default name exist."
    ),
    is_eager=True,
)
@click.option(
    "--auto-accept",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="Flag responsible for whether to skip confirmation of the generated config.",
)
@click.option(
    "--state-path",
    type=click.Path(file_okay=False, resolve_path=True, path_type=Path),
    callback=validate_state_path,
    # required=True,
    help="The path to the directory where the state path should be stored to.",
)
@click.option(
    "--raise-error-missing-reference",
    type=bool,
    default=False,
    show_default=True,
    help="Flag responsible for raising error if references are missing.",
)
@click.pass_context
def config_command(
    ctx: click.Context,
    config_path: Path,
    auto_accept: bool,
    state_path: Path,
    raise_error_missing_reference: bool,
) -> None:
    common_config, config_file = get_config_file_content(config_path)
    old_config = deepcopy(common_config)
    try:
        qualibrate_version_validator(common_config, False)
    except InvalidQualibrateConfigVersion:
        if common_config:
            migrate_command(["--config-path", config_path], standalone_mode=False)
            common_config, config_file = get_config_file_content(config_path)
    quam_config = common_config.get(QUAM_CONFIG_KEY, {})
    names_mapping = {
        "state_path": "state_path",
        "raise_error_missing_reference": "raise_error_missing_reference",
    }
    quam_config = get_config_by_args_mapping(
        names_mapping,
        quam_config,
        ctx,
    )
    quam_top_l = QuamTopLevelConfig({QUAM_CONFIG_KEY: quam_config})
    try:
        write_config(
            config_file,
            common_config,
            quam_top_l.quam,
            QUAM_CONFIG_KEY,
            quam_before_write_cb,
            confirm=not auto_accept,
        )
    except Exit:
        if old_config:
            simple_write(config_file, old_config)
        raise


def quam_before_write_cb(config: QuamConfig) -> None:
    if config.state_path:
        config.state_path.mkdir(parents=True, exist_ok=True)


if __name__ == "__main__":
    config_command(
        ["--raise-error-on-missing-references", "True"], standalone_mode=False
    )



from typing import Union

__all__ = ["type_is_optional"]


def type_is_optional(type_) -> bool:
    """Check if a type is Optional[T] for some type T.

    Args:
        type_: The type to check.

    Returns:
        True if the type is Optional[T] for some type T, False otherwise.

    Notes:
        This function does not check if the type is a Union of None and some other
        type, only if it is a Union of exactly two types, one of which is None.
        So Optional[Union[str, int]] will return False, but Optional[str] will return
        True.
    """
    if not hasattr(type_, "__origin__"):
        return False
    if type_.__origin__ is not Union:
        return False
    if len(type_.__args__) != 2:
        return False
    if type_.__args__[1] is not type(None):
        return False
    return True



from pathlib import Path
from typing import Optional

from qualibrate_config.models.base.config_base import BaseConfig


class SerializationConfig(BaseConfig):
    include_defaults: bool = True


class QuamConfig(BaseConfig):
    version: int = 3
    state_path: Optional[Path] = None
    raise_error_missing_reference: bool = False
    serialization: Optional[SerializationConfig] = None


class QuamTopLevelConfig(BaseConfig):
    quam: QuamConfig



from quam import QuamRoot, QuamComponent, quam_dataclass
import warnings


@quam_dataclass
class Component(QuamComponent):
    y: int = 2


@quam_dataclass
class Root(QuamRoot):
    a: Component
    b: Component = "#/a"


def test_quam_root_reference():
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        root = Root(a=Component())
        assert root.a is root.b


def test_quam_root_reference_to_dict():
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        root = Root(a="#/b", b=Component())
        d = root.to_dict(include_defaults=False)
        assert d == {
            "a": "#/b",
            "b": {"__class__": "test_quam_root_referencing.Component"},
            "__class__": "test_quam_root_referencing.Root",
        }



import json
import pytest
from quam.serialisation.json import JSONSerialiser


def test_save_backwards_compatibility_single_dict(
    serialiser, sample_quam_object, tmp_path
):
    """Test save still works with a single dict mapping (backward compatibility)."""
    target_folder = tmp_path / "single_dict_save"
    single_mapping = {"wiring_bc.json": ["wiring"]}

    serialiser.save(
        quam_obj=sample_quam_object,
        path=target_folder,
        content_mapping=single_mapping,
        include_defaults=False,
    )

    wiring_path = target_folder / "wiring_bc.json"
    default_path = target_folder / serialiser.default_filename

    assert wiring_path.exists()
    assert default_path.exists()
    with wiring_path.open("r", encoding="utf-8") as f:
        assert json.load(f) == {"wiring": sample_quam_object.wiring}
    with default_path.open("r", encoding="utf-8") as f:
        loaded = json.load(f)
        assert "components" in loaded
        assert loaded["other"] == "specific_value"
        assert loaded["__class__"] == "conftest.MockQuamRoot"


def test_validate_and_convert_content_mapping_new_format():
    """Test new format content mapping is correctly identified and validated."""
    # New format: component -> filename
    new_format = {"component1": "file1.json", "component2": "file2.json"}
    result = JSONSerialiser._validate_and_convert_content_mapping(new_format)
    assert result == new_format

    # String to string mapping should be identified as new format
    string_to_string = {"hi": "bye.json"}
    result = JSONSerialiser._validate_and_convert_content_mapping(string_to_string)
    assert result == string_to_string


def test_validate_and_convert_content_mapping_old_format():
    """Test old format content mapping is correctly identified and converted."""
    # Old format: filename -> [components]
    old_format_list = {"file.json": ["component1", "component2"]}
    expected = {"component1": "file.json", "component2": "file.json"}
    result = JSONSerialiser._validate_and_convert_content_mapping(old_format_list)
    assert result == expected

    # Test with tuple
    old_format_tuple = {"file.json": ("component1", "component2")}
    result = JSONSerialiser._validate_and_convert_content_mapping(old_format_tuple)
    assert result == expected

    # Test with set
    old_format_set = {"file.json": {"component1", "component2"}}
    result = JSONSerialiser._validate_and_convert_content_mapping(old_format_set)
    # Sets are unordered, so just check keys
    assert set(result.keys()) == {"component1", "component2"}
    assert all(value == "file.json" for value in result.values())


def test_validate_and_convert_content_mapping_conflicts():
    """Test handling of conflicts in old format conversion."""
    # Components assigned to multiple files
    conflicting_mapping = {
        "file1.json": ["component1", "component2"],
        "file2.json": ["component2", "component3"],  # component2 appears twice
    }
    result = JSONSerialiser._validate_and_convert_content_mapping(conflicting_mapping)
    # component2 should get the last assignment (file2.json)
    assert result["component1"] == "file1.json"
    assert result["component2"] == "file2.json"
    assert result["component3"] == "file2.json"


def test_validate_and_convert_content_mapping_invalid():
    """Test invalid content mapping formats."""
    # Mixed value types
    with pytest.raises(TypeError):
        JSONSerialiser._validate_and_convert_content_mapping(
            {"file1.json": ["component1"], "component2": "file2.json"}
        )

    # Invalid key in old format
    with pytest.raises(TypeError):
        JSONSerialiser._validate_and_convert_content_mapping({123: ["component1"]})

    # Invalid value in old format
    with pytest.raises(TypeError):
        JSONSerialiser._validate_and_convert_content_mapping({"file.json": [123]})

    # Invalid key in new format
    with pytest.raises(TypeError):
        JSONSerialiser._validate_and_convert_content_mapping({123: "file.json"})

    # Invalid value in new format
    with pytest.raises(TypeError):
        JSONSerialiser._validate_and_convert_content_mapping({"component": 123})


def test_validate_and_convert_content_mapping_empty():
    """Test handling of empty content mapping."""
    assert JSONSerialiser._validate_and_convert_content_mapping({}) == {}
    assert JSONSerialiser._validate_and_convert_content_mapping(None) == {}



from copy import deepcopy
import pytest
from unittest.mock import patch

from quam.core import *


@pytest.fixture
def BareQuamRoot():
    @quam_dataclass
    class BareQuamRoot(QuamRoot): ...

    return BareQuamRoot


@pytest.fixture
def BareQuamComponent():
    @quam_dataclass
    class BareQuamComponent(QuamComponent): ...

    return BareQuamComponent


@pytest.fixture(scope="function", autouse=True)
def remove_quam_root():
    from quam.core import QuamBase

    QuamBase._last_instantiated_root = None


@pytest.fixture
def qua_config():
    from quam.core.qua_config_template import qua_config_template

    return deepcopy(qua_config_template)


@pytest.fixture(scope="session", autouse=True)
def prevent_default_config_loading():
    """
    Prevents loading ~/.qualibrate config during tests by changing the default
    config filename to a non-existent file.

    This ensures tests pass regardless of whether ~/.qualibrate exists on the system.
    """
    try:
        import qualibrate_config.vars
        original_filename = qualibrate_config.vars.DEFAULT_CONFIG_FILENAME
        new_filename = "nonexisting_config_for_testing.toml"
        qualibrate_config.vars.DEFAULT_CONFIG_FILENAME = new_filename
        yield
        qualibrate_config.vars.DEFAULT_CONFIG_FILENAME = original_filename
    except ImportError:
        # qualibrate_config not available, no need to mock
        yield


@pytest.fixture(scope="function", autouse=True)
def mock_quam_config():
    """
    Mock get_quam_config to prevent loading system config during tests.

    Patches all import locations where get_quam_config is used to ensure
    no config files are loaded from the system during testing.

    Returns a minimal mock config object with default values matching QuamConfig.
    """
    # Create a minimal mock config object that matches QuamConfig structure
    class MockSerializationConfig:
        include_defaults = True

    class MockQuamConfig:
        version = 3
        state_path = None
        raise_error_missing_reference = False
        serialization = MockSerializationConfig()

    mock_config = MockQuamConfig()

    import_locations = [
        'quam.core.quam_classes.get_quam_config',
        'quam.serialisation.json.get_quam_config',
        'quam.components.get_quam_config',
    ]

    patches = [patch(location) for location in import_locations]
    mocks = []

    for p in patches:
        mock = p.start()
        mock.return_value = mock_config
        mocks.append(mock)

    yield mocks

    for p in patches:
        p.stop()



from typing import Dict
from dataclasses import field
import warnings
from quam.core import QuamRoot, quam_dataclass
from quam.components.channels import Channel
from quam.components.octave import Octave
from quam.components.ports import FEMPortsContainer, OPXPlusPortsContainer

__all__ = [
    "BasicQuam",
    "BasicFEMQuam",
    "BasicOPXPlusQuam",
]


@quam_dataclass
class BasicQuam(QuamRoot):
    """Basic top-level QUAM root component.

    If custom QUAM components are used, a custom QUAM root component should be created.

    Args:
        channels (Dict[str, Channel], optional): A dictionary of channels.
        octaves (Dict[str, Octave], optional): A dictionary of octaves.
    """

    channels: Dict[str, Channel] = field(default_factory=dict)
    octaves: Dict[str, Octave] = field(default_factory=dict)


@quam_dataclass
class BasicFEMQuam(BasicQuam):
    """Basic QUAM root component for FEM (Frequency Encoded Multiplexing) systems.

    Extends BasicQuam with FEM-specific port configurations for quantum control hardware.

    Attributes:
        ports (FEMPortsContainer): Container for FEM-specific analog and digital ports.
    """

    ports: FEMPortsContainer = field(default_factory=FEMPortsContainer)


@quam_dataclass
class BasicOPXPlusQuam(BasicQuam):
    """Basic QUAM root component for OPX+ systems.

    Extends BasicQuam with OPX+-specific port configurations for quantum control hardware.

    Attributes:
        ports (OPXPlusPortsContainer): Container for OPX+ analog and digital ports.
    """

    ports: OPXPlusPortsContainer = field(default_factory=OPXPlusPortsContainer)


class BasicQuAM(BasicQuam):
    def __post_init__(self, *args, **kwargs):
        warnings.warn(
            "BasicQuAM is deprecated. Use BasicQuam instead.",
            DeprecationWarning,
        )
        super().__post_init__(*args, **kwargs)



from quam.core.macro.method_macro import MethodMacro, method_macro


class TestClass:
    def __init__(self, value: int):
        self.value = value

    @method_macro
    def add(self, x: int) -> int:
        return self.value + x


def test_method_macro_binding():
    """Test that MethodMacro correctly binds to instance methods"""
    obj = TestClass(5)
    assert isinstance(obj.add, MethodMacro)
    assert MethodMacro.is_macro_method(obj.add)
    assert obj.add.instance == obj


def test_method_macro_apply():
    """Test that MethodMacro.apply works with instance methods"""
    obj = TestClass(5)
    assert obj.add.apply(3) == 8  # 5 + 3
    assert obj.add(3) == 8  # Should work the same way


def test_is_macro_method():
    """Test the is_macro_method static method"""
    obj = TestClass(5)

    assert MethodMacro.is_macro_method(obj.add)
    assert not MethodMacro.is_macro_method(lambda x: x)
    assert not MethodMacro.is_macro_method(42)


def test_method_macro_preserves_metadata():
    """Test that MethodMacro preserves the original function's metadata"""

    def original(x: int) -> int:
        """Test docstring"""
        return x

    decorated = MethodMacro(original)
    assert decorated.__doc__ == original.__doc__
    assert decorated.__name__ == original.__name__



import numpy as np
from typing import List

from quam.core import QuamComponent, quam_dataclass
from quam.utils import string_reference as str_ref


__all__ = [
    "LocalOscillator",
    "Mixer",
    "FrequencyConverter",
]


@quam_dataclass
class LocalOscillator(QuamComponent):
    """QUAM component for a local oscillator.

    Args:
        frequency (float): The frequency of the local oscillator.
            Used by the mixer to determine the intermediate frequency.
        power (float, optional): The power of the local oscillator.
            Not used for the QUA configuration
    """

    frequency: float = None
    power: float = None

    def configure(self): ...


@quam_dataclass
class Mixer(QuamComponent):
    """QUAM component for a mixer.

    All properties are optional, so it can be instantiated as `Mixer()`.
    For the default values, it is assumed that the mixer parent is an `IQChannel`
    that has a `LocalOscillator`.

    Args:
        local_oscillator_frequency (float, optional): The frequency of the local
            oscillator. Default is `#../local_oscillator/frequency`, meaning that
            the frequency is extracted from the the local_oscillator of the parent.
        intermediate_frequency (float, optional): The intermediate frequency of the
            mixer. Default is `#../intermediate_frequency`, meaning that the frequency
            references the intermediate_frequency of the parent.
        correction_gain (float, optional): The gain imbalance of the mixer.
            Default is 0, see `Mixer.IQ_imbalance` for details.
        correction_phase (float, optional): The phase imbalance of the mixer in radians.
    """

    local_oscillator_frequency: float = "#../local_oscillator/frequency"
    intermediate_frequency: float = "#../../intermediate_frequency"

    correction_gain: float = 0
    correction_phase: float = 0

    @property
    def name(self):
        frequency_converter = getattr(self, "parent", None)
        if frequency_converter is None:
            raise AttributeError(
                f"Mixer.parent must be a frequency converter for {self}"
            )

        channel = getattr(frequency_converter, "parent", None)
        channel_name = getattr(channel, "name", None)
        if channel is None or channel_name is None:
            raise AttributeError(f"Mixer.parent.parent must be a channel for {self}")

        return f"{channel_name}{str_ref.DELIMITER}mixer"

    def apply_to_config(self, config: dict):
        """Adds this mixer to the QUA configuration.

        See [`QuamComponent.apply_to_config`][quam.core.quam_classes.QuamComponent.apply_to_config]
        for details.
        """
        correction_matrix = self.IQ_imbalance(
            self.correction_gain, self.correction_phase
        )

        config["mixers"][self.name] = [
            {
                "intermediate_frequency": self.intermediate_frequency,
                "lo_frequency": self.local_oscillator_frequency,
                "correction": correction_matrix,
            }
        ]

    @staticmethod
    def IQ_imbalance(g: float, phi: float) -> List[float]:
        """
        Creates the correction matrix for the mixer imbalance caused by the gain and
        phase imbalances, more information can be seen here:
        https://docs.qualang.io/libs/examples/mixer-calibration/#non-ideal-mixer
        :param g: relative gain imbalance between the I & Q ports. (unit-less),
            set to 0 for no gain imbalance.
        :param phi: relative phase imbalance between the I & Q ports (radians),
            set to 0 for no phase imbalance.
        """
        c = np.cos(phi)
        s = np.sin(phi)
        N = 1 / ((1 - g**2) * (2 * c**2 - 1))
        return [
            float(N * x) for x in [(1 - g) * c, (1 + g) * s, (1 - g) * s, (1 + g) * c]
        ]


@quam_dataclass
class BaseFrequencyConverter(QuamComponent):
    """Base class for frequency converters."""

    pass


@quam_dataclass
class FrequencyConverter(BaseFrequencyConverter):
    """Frequency up/down converter component.

    This component encapsulates the local oscillator and mixer used to upconvert or
    downconvert an RF signal.

    The FrequencyConverter component is attached to IQ channels through

    - `IQChannel.frequency_converter_up`
    - `InOutIQChannel.frequency_converter_down`

    Args:
        local_oscillator (LocalOscillator): The local oscillator for the frequency converter.
        mixer (Mixer): The mixer for the frequency converter.
        gain (float): The gain of the frequency converter.
    """

    local_oscillator: LocalOscillator = None
    mixer: Mixer = None
    gain: float = None

    @property
    def LO_frequency(self):
        return self.local_oscillator.frequency

    @LO_frequency.setter
    def LO_frequency(self, value):
        """Sets the frequency of the local oscillator object"""
        if self.local_oscillator is None:
            raise AttributeError(
                f"Unable to set LO frequency for {self} as it has no local oscillator"
            )

        # Use set_at_reference to ensure the frequency is updated, even if the local
        # oscillator frequency is a reference
        self.local_oscillator.set_at_reference("frequency", value)

    def configure(self):
        if self.local_oscillator is not None:
            self.local_oscillator.configure()



import warnings
from abc import abstractclassmethod


instantiation_deprecations = []


class InstantiationDeprecationRule:
    @abstractclassmethod
    def match(cls, quam_class, contents):
        raise NotImplementedError

    @abstractclassmethod
    def apply(cls, quam_class, contents):
        raise NotImplementedError


class DeprecatedFrequencyConverterInstantiation(InstantiationDeprecationRule):
    @classmethod
    def match(cls, quam_class, contents):
        from quam.components.hardware import BaseFrequencyConverter

        if quam_class != BaseFrequencyConverter:
            return False
        if "__class__" in contents:
            return False
        return True

    @classmethod
    def apply(cls, quam_class, contents):
        from quam.components.hardware import FrequencyConverter

        warnings.warn(
            "The default frequency converter for channels is changed to the "
            "`BaseFrequencyConverter`. If you want to use `FrequencyConverter`, "
            'Please add {"__class__": "quam.components.hardware.FrequencyConverter"} '
            "to the JSON contents of the frequency converter. This will raise an error "
            "in future versions.",
            DeprecationWarning,
        )
        contents["__class__"] = "quam.components.hardware.FrequencyConverter"

        return FrequencyConverter, contents


instantiation_deprecations.append(DeprecatedFrequencyConverterInstantiation)



from .analog_outputs import *
from .analog_inputs import *
from .base_ports import *
from .digital_inputs import *
from .digital_outputs import *
from .ports_containers import *

__all__ = [
    *analog_outputs.__all__,
    *analog_inputs.__all__,
    *base_ports.__all__,
    *digital_inputs.__all__,
    *digital_outputs.__all__,
    *ports_containers.__all__,
]



from collections import UserDict
from dataclasses import field
from typing import Dict, List
import pytest
from quam.components import Qubit, QubitPair
from quam.components.channels import IQChannel
from quam.core.quam_classes import QuamRoot, quam_dataclass


@quam_dataclass
class MockQubit(Qubit):
    xy: IQChannel = None


@quam_dataclass
class MockQubitPair(QubitPair):
    qubit_control: MockQubit
    qubit_target: MockQubit


@quam_dataclass
class Quam(QuamRoot):
    qubits: Dict[str, MockQubit]
    qubit_pairs: Dict[str, MockQubitPair] = field(default_factory=dict)


@pytest.fixture
def test_qubit_control():
    return MockQubit(
        id="q1",
        xy=IQChannel(
            id="xy_control",
            opx_output_I=("con1", 1),
            opx_output_Q=("con1", 2),
            frequency_converter_up=None,
        ),
    )


@pytest.fixture
def test_qubit_target():
    return MockQubit(
        id="q2",
        xy=IQChannel(
            id="xy_target",
            opx_output_I=("con1", 5),
            opx_output_Q=("con1", 6),
            frequency_converter_up=None,
        ),
    )


@pytest.fixture
def test_quam(test_qubit_control, test_qubit_target):
    machine = Quam(
        qubits={"control": test_qubit_control, "target": test_qubit_target},
    )

    machine.qubit_pairs["pair_1"] = MockQubitPair(
        id="pair_1",
        qubit_control=test_qubit_control.get_reference(),
        qubit_target=test_qubit_target.get_reference(),
    )
    return machine


@pytest.fixture
def test_qubit_pair(test_quam):
    return test_quam.qubit_pairs["pair_1"]


def test_qubit_pair_initialization(
    test_qubit_pair, test_qubit_control, test_qubit_target
):
    """Test that QubitPair is initialized correctly"""
    assert test_qubit_pair.qubit_control == test_qubit_control
    assert test_qubit_pair.qubit_target == test_qubit_target
    assert test_qubit_pair.name == "pair_1"
    assert isinstance(test_qubit_pair.macros, UserDict)
    assert len(test_qubit_pair.macros) == 0


def test_qubit_pair_align(test_qubit_pair, mocker):
    """Test that align method calls the control qubit's align method with correct args"""
    mock_align = mocker.patch.object(test_qubit_pair.qubit_control, "align")

    test_qubit_pair.align()

    mock_align.assert_called_once_with(test_qubit_pair.qubit_target)


def test_qubit_pair_via_matmul(test_quam):
    """Test that qubit pair can be accessed via @ operator"""
    control = test_quam.qubits["control"]
    target = test_quam.qubits["target"]

    qubit_pair = control @ target

    assert isinstance(qubit_pair, QubitPair)
    assert qubit_pair.qubit_control == control
    assert qubit_pair.qubit_target == target


def test_matmul_with_invalid_qubit(test_quam):
    """Test that @ operator raises error for invalid qubit pairs"""
    control = test_quam.qubits["control"]

    with pytest.raises(ValueError, match="Cannot create a qubit pair with same qubit"):
        _ = control @ control

    with pytest.raises(
        ValueError, match="Cannot create a qubit pair .* with a non-qubit object"
    ):
        _ = control @ "not_a_qubit"


def test_matmul_with_nonexistent_pair(test_quam):
    """Test that @ operator raises error for non-existent qubit pairs"""
    target = test_quam.qubits["target"]
    control = test_quam.qubits["control"]

    # Try to access pair in reverse order (target @ control) when only (control @ target) exists
    with pytest.raises(ValueError, match="Qubit pair not found"):
        _ = target @ control



from quam.components.quantum_components.quantum_component import QuantumComponent
from quam.components.quantum_components.qubit import Qubit
from quam.components.quantum_components.qubit_pair import QubitPair

__all__ = ["QuantumComponent", "Qubit", "QubitPair"]



import pytest
from quam.components import IQChannel
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort


def test_IQ_channel_set_dc_offset(mocker):
    mocker.patch("quam.components.channels.set_dc_offset")

    channel = IQChannel(
        id="channel",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
    )

    with pytest.raises(TypeError):
        channel.set_dc_offset(0.5)

    with pytest.raises(ValueError):
        channel.set_dc_offset(0.5, "X")

    channel.set_dc_offset(0.5, "I")

    from quam.components.channels import set_dc_offset

    set_dc_offset.assert_called_once_with(
        element="channel", element_input="I", offset=0.5
    )


def test_IQ_channel_inferred_RF_frequency():
    channel = IQChannel(
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
    )

    assert channel.intermediate_frequency is None
    assert channel.LO_frequency == "#./frequency_converter_up/LO_frequency"
    assert channel.RF_frequency == "#./inferred_RF_frequency"
    with pytest.raises(AttributeError):
        channel.inferred_RF_frequency()

    channel.LO_frequency = None
    channel.LO_frequency = 5e9
    channel.intermediate_frequency = 100e6
    assert channel.inferred_RF_frequency == 5.1e9


def test_IQ_channel_inferred_intermediate_frequency():
    channel = IQChannel(
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
        intermediate_frequency="#./inferred_intermediate_frequency",
        LO_frequency=5.1e9,
        RF_frequency=5.2e9,
    )

    assert channel.intermediate_frequency == 100e6

    channel.LO_frequency = None
    with pytest.raises(AttributeError):
        channel.inferred_intermediate_frequency

    channel.LO_frequency = 5.1e9
    channel.RF_frequency = None
    with pytest.raises(AttributeError):
        channel.inferred_intermediate_frequency


def test_IQ_channel_inferred_LO_frequency():
    channel = IQChannel(
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
        intermediate_frequency=100e6,
        LO_frequency="#./inferred_LO_frequency",
        RF_frequency=5.2e9,
    )

    assert channel.LO_frequency == 5.1e9

    channel.RF_frequency = None
    with pytest.raises(AttributeError):
        channel.inferred_LO_frequency

    channel.RF_frequency = 5.2e9
    channel.intermediate_frequency = None
    with pytest.raises(AttributeError):
        channel.inferred_LO_frequency


def test_generate_config(qua_config):
    channel = IQChannel(
        id="out_channel",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
    )

    channel.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_outputs": {
                1: {"delay": 0, "shareable": False},
                2: {"delay": 0, "shareable": False},
            },
        }
    }

    assert qua_config["elements"] == {
        "out_channel": {
            "mixInputs": {
                "I": ("con1", 1),
                "Q": ("con1", 2),
            },
            "operations": {},
        }
    }


def test_generate_config_ports(qua_config):
    channel = IQChannel(
        id="out_channel",
        opx_output_I=OPXPlusAnalogOutputPort("con1", 1),
        opx_output_Q=OPXPlusAnalogOutputPort("con1", 2),
        frequency_converter_up=None,
    )

    channel.apply_to_config(qua_config)
    channel.opx_output_I.apply_to_config(qua_config)
    channel.opx_output_Q.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_outputs": {
                1: {"delay": 0, "shareable": False},
                2: {"delay": 0, "shareable": False},
            },
        }
    }

    assert qua_config["elements"] == {
        "out_channel": {
            "mixInputs": {
                "I": ("con1", 1),
                "Q": ("con1", 2),
            },
            "operations": {},
        }
    }



import pytest
import json


def test_save_split_content_basic(serialiser, sample_quam_object, tmp_path):
    """Test splitting content into specified files."""
    folder = tmp_path / "split_basic"
    # Use the QuamRoot's to_dict method
    full_contents = sample_quam_object.to_dict(include_defaults=False)

    # Expected:
    # {'components': {'comp1': {'id': 'c1'}, 'comp2': {'id': 'c2', 'value': 5}},
    #  'wiring': {'w1': 'p1', 'w2': 'p2'},
    #  'other': 'specific_value'}

    mapping = {"wiring": "wiring.json", "components": "components.json"}

    serialiser._save_split_content(full_contents, folder, mapping)

    wiring_path = folder / "wiring.json"
    components_path = folder / "components.json"
    default_path = folder / serialiser.default_filename  # state.json

    assert wiring_path.exists()
    assert components_path.exists()
    assert default_path.exists()  # 'other' remains

    with wiring_path.open("r", encoding="utf-8") as f:
        assert json.load(f) == {"wiring": {"w1": "p1", "w2": "p2"}}
    with components_path.open("r", encoding="utf-8") as f:
        # Check against the dict representation of the components part
        expected_components = {
            "components": {
                "comp1": {
                    "id": "c1",
                    "__class__": "conftest.MockMainComponent",
                },  # value=1 is default
                "comp2": {
                    "id": "c2",
                    "value": 5,
                    "__class__": "conftest.MockMainComponent",
                },  # value=5 is not default
            }
        }
        d = json.load(f)
        assert d == expected_components
    with default_path.open("r", encoding="utf-8") as f:
        assert json.load(f) == {
            "other": "specific_value",
            "__class__": "conftest.MockQuamRoot",
        }  # default_val is default


def test_save_split_content_missing_keys_warning(
    serialiser, sample_quam_object, tmp_path
):
    """Test warning when specified keys are missing."""
    folder = tmp_path / "split_missing"
    full_contents = sample_quam_object.to_dict(include_defaults=False)
    mapping = {"nonexistent_key": "missing.json", "wiring": "missing.json"}

    with pytest.warns(UserWarning, match="specified in content_mapping was not found"):
        serialiser._save_split_content(full_contents, folder, mapping)

    missing_path = folder / "missing.json"
    default_path = folder / serialiser.default_filename

    assert missing_path.exists()  # Should still create file for 'wiring'
    assert default_path.exists()  # 'components' and 'other' remain

    with missing_path.open("r", encoding="utf-8") as f:
        assert json.load(f) == {
            "wiring": {"w1": "p1", "w2": "p2"}
        }  # Only wiring is saved
    with default_path.open("r", encoding="utf-8") as f:
        loaded_default = json.load(f)
        assert "components" in loaded_default
        assert "other" in loaded_default


def test_save_split_content_no_remaining(serialiser, sample_quam_object, tmp_path):
    """Test when splitting consumes all content, no default file needed."""
    folder = tmp_path / "split_no_remaining"
    full_contents = sample_quam_object.to_dict(include_defaults=False)
    # Map all keys present when not including defaults
    mapping = {
        "components": "comps.json",
        "wiring": "wires.json",
        "other": "misc.json",
    }

    serialiser._save_split_content(full_contents, folder, mapping)

    default_path = folder / serialiser.default_filename

    assert (folder / "comps.json").exists()
    assert (folder / "wires.json").exists()
    assert (folder / "misc.json").exists()
    with (folder / "misc.json").open("r", encoding="utf-8") as f:
        assert json.load(f) == {
            "other": "specific_value",
        }
    with default_path.open("r", encoding="utf-8") as f:
        assert json.load(f) == {
            "__class__": "conftest.MockQuamRoot",
        }


def test_save_split_content_create_folder(serialiser, sample_quam_object, tmp_path):
    """Test that the target folder is created if it doesn't exist."""
    folder = tmp_path / "new_dir_split"
    assert not folder.exists()
    full_contents = sample_quam_object.to_dict(include_defaults=False)
    mapping = {"wiring": "wiring.json"}

    serialiser._save_split_content(full_contents, folder, mapping)

    assert folder.exists()
    assert (folder / "wiring.json").exists()
    assert (folder / serialiser.default_filename).exists()


def test_save_split_content_create_subfolder(serialiser, sample_quam_object, tmp_path):
    """Test creating subdirectories specified in the mapping filename."""
    folder = tmp_path / "split_subfolder"
    full_contents = sample_quam_object.to_dict(include_defaults=False)
    subfolder_name = "sub"
    mapping = {
        "wiring": f"{subfolder_name}/wiring.json",
        "components": "components.json",
    }

    serialiser._save_split_content(full_contents, folder, mapping)

    subfolder_path = folder / subfolder_name
    wiring_path = subfolder_path / "wiring.json"
    components_path = folder / "components.json"
    default_path = folder / serialiser.default_filename

    assert subfolder_path.exists()
    assert subfolder_path.is_dir()
    assert wiring_path.exists()
    assert components_path.exists()
    assert default_path.exists()

    with wiring_path.open("r", encoding="utf-8") as f:
        assert "wiring" in json.load(f)
    with components_path.open("r", encoding="utf-8") as f:
        assert "components" in json.load(f)
    with default_path.open("r", encoding="utf-8") as f:
        # Includes 'other' and default 'default_val'
        loaded_default = json.load(f)
        assert loaded_default == {
            "other": "specific_value",
            "__class__": "conftest.MockQuamRoot",
        }


def test_save_split_content_absolute_path_warning(
    serialiser, sample_quam_object, tmp_path
):
    """Test warning when an absolute path is used in content_mapping."""
    folder = tmp_path / "split_abs_warn"
    abs_path_str = str((tmp_path / "ignored_abs_path" / "wiring.json").absolute())
    full_contents = sample_quam_object.to_dict(include_defaults=False)
    mapping = {"wiring": abs_path_str}

    with pytest.warns(
        UserWarning, match="Absolute path .* in content_mapping is ignored"
    ):
        serialiser._save_split_content(full_contents, folder, mapping)

    assert (folder / "wiring.json").exists()
    assert not (tmp_path / "ignored_abs_path").exists()

    with (folder / "wiring.json").open("r", encoding="utf-8") as f:
        assert "wiring" in json.load(f)


def test_save_split_content_empty_input_no_mapping(serialiser, tmp_path):
    """Test saving completely empty content with no mapping does not save anything."""
    folder = tmp_path / "split_empty_input"
    full_contents = {}
    mapping = {}

    serialiser._save_split_content(full_contents, folder, mapping)

    assert not folder.exists()


def test_save_single_file(serialiser, sample_quam_object, tmp_path):
    """Test saving to a single JSON file when path has .json suffix."""
    filepath = tmp_path / "single_save.json"
    serialiser.save(sample_quam_object, filepath, include_defaults=False)

    assert filepath.exists()
    with filepath.open("r", encoding="utf-8") as f:
        loaded = json.load(f)
    # Compare with dict representation (explicitly exclude defaults)
    expected_dict = sample_quam_object.to_dict(include_defaults=False)
    expected_dict["__class__"] = (  # Class added by default for root
        f"{sample_quam_object.__module__}.{sample_quam_object.__class__.__name__}"
    )
    assert loaded == expected_dict


def test_save_single_file_include_defaults(
    serialiser_with_defaults, sample_quam_object, tmp_path
):
    """Test saving to a single JSON file including default values."""
    filepath = tmp_path / "single_save_defaults.json"
    serialiser_with_defaults.save(sample_quam_object, filepath)

    assert filepath.exists()
    with filepath.open("r", encoding="utf-8") as f:
        loaded = json.load(f)
    # Compare with dict representation including defaults
    expected_dict = sample_quam_object.to_dict(include_defaults=True)
    expected_dict["__class__"] = (
        f"{sample_quam_object.__module__}.{sample_quam_object.__class__.__name__}"
    )
    assert loaded == expected_dict


def test_save_split_via_save_method(serialiser, sample_quam_object, tmp_path):
    """Test saving split content using the main save method (path is a folder)."""
    folder = tmp_path / "save_split_folder"
    mapping = {"wiring.json": ["wiring"]}
    serialiser.save(sample_quam_object, folder, content_mapping=mapping)

    assert folder.exists()
    assert folder.is_dir()
    assert (folder / "wiring.json").exists()
    assert (folder / serialiser.default_filename).exists()  # components and other

    with (folder / "wiring.json").open("r", encoding="utf-8") as f:
        assert "wiring" in json.load(f)
    with (folder / serialiser.default_filename).open("r", encoding="utf-8") as f:
        loaded_default = json.load(f)
        assert "components" in loaded_default
        assert "other" in loaded_default
        # __class__ should also be in the default file if not split out
        assert "__class__" in loaded_default


def test_save_with_ignore(serialiser, sample_quam_object, tmp_path):
    """Test save method with 'ignore' argument for single file."""
    filepath = tmp_path / "save_ignore.json"
    ignore_keys = ["wiring", "default_val"]
    serialiser.save(sample_quam_object, filepath, ignore=ignore_keys, include_defaults=False)

    assert filepath.exists()
    with filepath.open("r", encoding="utf-8") as f:
        loaded = json.load(f)

    expected_dict = sample_quam_object.to_dict(include_defaults=False)
    expected_dict.pop("wiring", None)
    expected_dict.pop(
        "default_val", None
    )  # Will not be present anyway if include_defaults=False
    expected_dict["__class__"] = (
        f"{sample_quam_object.__module__}.{sample_quam_object.__class__.__name__}"
    )

    assert loaded == expected_dict


def test_save_split_with_ignore(serialiser, sample_quam_object, tmp_path):
    """Test save method with 'ignore' argument for split content."""
    folder = tmp_path / "save_split_ignore"
    mapping = {"components.json": ["components"]}
    ignore_keys = ["wiring"]

    serialiser.save(
        sample_quam_object, folder, content_mapping=mapping, ignore=ignore_keys
    )

    assert folder.exists()
    assert (folder / "components.json").exists()
    assert (folder / serialiser.default_filename).exists()  # 'other' + __class__
    assert not (folder / "wiring.json").exists()  # Ignored

    with (folder / "components.json").open("r", encoding="utf-8") as f:
        assert "components" in json.load(f)
    with (folder / serialiser.default_filename).open("r", encoding="utf-8") as f:
        loaded_default = json.load(f)
        assert "other" in loaded_default
        assert "__class__" in loaded_default
        assert "wiring" not in loaded_default


def test_save_unsupported_path_suffix(serialiser, sample_quam_object, tmp_path):
    """Test ValueError for unsupported file extensions."""
    filepath = tmp_path / "unsupported.txt"
    with pytest.raises(ValueError, match="Unsupported path suffix"):
        serialiser.save(sample_quam_object, filepath)


def test_save_default_path(serialiser, sample_quam_object, monkeypatch, tmp_path):
    """Test saving to default path (mocking _get_state_path)."""
    default_save_path = tmp_path / "default_state_dir"
    monkeypatch.setattr(serialiser, "_get_state_path", lambda: default_save_path)

    # Default mapping is empty, save goes to default_filename inside folder
    serialiser.save(sample_quam_object, include_defaults=False)

    expected_file = default_save_path / serialiser.default_filename
    assert default_save_path.exists()
    assert expected_file.exists()

    with expected_file.open("r", encoding="utf-8") as f:
        loaded = json.load(f)
    expected_dict = sample_quam_object.to_dict(include_defaults=False)
    expected_dict["__class__"] = (
        f"{sample_quam_object.__module__}.{sample_quam_object.__class__.__name__}"
    )
    assert loaded == expected_dict


def test_save_nested_object(serialiser, sample_quam_object_nested, tmp_path):
    """Test saving an object with properly nested QuamComponents."""
    filepath = tmp_path / "nested_save.json"
    serialiser.save(
        sample_quam_object_nested, filepath, include_defaults=False
    )  # Exclude defaults

    assert filepath.exists()
    with filepath.open("r", encoding="utf-8") as f:
        loaded = json.load(f)

    # Check structure - exclude defaults for comparison
    expected = sample_quam_object_nested.to_dict(include_defaults=False)
    expected["__class__"] = (
        f"{sample_quam_object_nested.__module__}.{sample_quam_object_nested.__class__.__name__}"
    )
    # Manually add __class__ for nested components if needed by loader (though save doesn't add it by default unless types mismatch)
    # expected["components"]["parent"]["__class__"] = f"{sample_quam_object_nested.components['parent'].__module__}.{sample_quam_object_nested.components['parent'].__class__.__name__}"
    # expected["components"]["parent"]["child"]["__class__"] = f"{sample_quam_object_nested.components['parent'].child.__module__}.{sample_quam_object_nested.components['parent'].child.__class__.__name__}"

    assert loaded == expected
    assert "child" in loaded["components"]["parent"]
    assert loaded["components"]["parent"]["child"]["id"] == "child1"
    # Check non-default values were saved
    assert loaded["components"]["parent"]["child"]["child_value"] == 99
    assert loaded["default_val"] == 15



from quam.utils.pulse import pulse_str_to_axis_axis_angle


def test_pulse_str_to_axis_angle():
    axis, angle = pulse_str_to_axis_axis_angle("X90")
    assert axis == "X"
    assert angle == 90

    axis, angle = pulse_str_to_axis_axis_angle("Xm90")
    assert axis == "X"
    assert angle == -90

    axis, angle = pulse_str_to_axis_axis_angle("X-90")
    assert axis == "X"
    assert angle == -90



import json
from typing import Dict, Optional
import pytest

from quam.serialisation import JSONSerialiser
from quam.core import QuamRoot, QuamComponent, quam_dataclass


@quam_dataclass
class Quam(QuamRoot):
    a: int
    b: list
    c: Optional[int] = None


def test_serialise_random_object(tmp_path):
    quam_root = Quam(a=1, b=[1, 2, 3])

    serialiser = JSONSerialiser()
    path = tmp_path / "quam_root.json"
    serialiser.save(quam_root, path, include_defaults=False)

    d = json.loads(path.read_text())

    assert d == {
        "a": 1,
        "b": [1, 2, 3],
        "__class__": "test_json_serialisation.Quam",
    }

    class RandomObject:
        pass

    quam_root.b = RandomObject()  # type: ignore

    with pytest.raises(TypeError):
        serialiser.save(quam_root, path, include_defaults=False)


def test_serialise_ignore(tmp_path):
    quam_root = Quam(a=1, b=[1, 2, 3])

    serialiser = JSONSerialiser()
    path = tmp_path / "quam_root.json"
    serialiser.save(quam_root, path, ignore=["b"], include_defaults=False)

    d = json.loads(path.read_text())

    assert d == {
        "a": 1,
        "__class__": "test_json_serialisation.Quam",
    }


def test_serialise_ignore_nonexisting(tmp_path):
    quam_root = Quam(a=1, b=[1, 2, 3])

    serialiser = JSONSerialiser()
    path = tmp_path / "quam_root.json"
    serialiser.save(quam_root, path, ignore=["c"], include_defaults=False)

    d = json.loads(path.read_text())

    assert d == {
        "a": 1,
        "b": [1, 2, 3],
        "__class__": "test_json_serialisation.Quam",
    }


@quam_dataclass
class Component(QuamComponent):
    a: int


def test_component_mapping(tmp_path):
    quam_root = Quam(a=1, b=Component(a=3), c=Component(a=4))

    serialiser = JSONSerialiser()
    path = tmp_path
    serialiser.save(quam_root, path, content_mapping={"b": "b.json"})

    d = json.loads((tmp_path / "b.json").read_text())
    assert d == {
        "b": {
            "__class__": "test_json_serialisation.Component",
            "a": 3,
        }
    }


def test_component_mapping_ignore(tmp_path):
    assert not (tmp_path / "b.json").exists()

    quam_root = Quam(a=1, b=Component(a=3), c=Component(a=4))

    serialiser = JSONSerialiser()
    serialiser.save(quam_root, tmp_path, ignore=["b"], content_mapping={"b": "b.json"})
    assert not (tmp_path / "b.json").exists()

    serialiser.save(
        quam_root,
        tmp_path,
        ignore=["b"],
        content_mapping={"b": "b.json", "c": "b.json"},
    )
    d = json.loads((tmp_path / "b.json").read_text())
    assert d == {
        "c": {
            "__class__": "test_json_serialisation.Component",
            "a": 4,
        }
    }


@quam_dataclass
class QuamWithIntDict(QuamRoot):
    a: int
    d: Dict[int, str]


def test_serialise_int_dict_keys(tmp_path):
    quam_root = QuamWithIntDict(a=1, d={1: "a", 2: "b"})

    serialiser = JSONSerialiser()
    path = tmp_path / "quam_root.json"
    serialiser.save(quam_root, path)

    d, _ = serialiser.load(path)

    assert d == {
        "a": 1,
        "d": {
            1: "a",
            2: "b",
        },
        "__class__": "test_json_serialisation.QuamWithIntDict",
    }



from typing import List
from dataclasses import is_dataclass, field
import pytest

from quam.core import QuamBase, QuamRoot, QuamComponent, QuamDict, quam_dataclass


def test_error_create_base_classes_directly():
    for cls in [QuamBase, QuamRoot, QuamComponent]:
        with pytest.raises(TypeError) as exc_info:
            cls()
        error_message = (
            f"Cannot instantiate {cls.__name__} directly. Please create a subclass and"
            " make it a dataclass."
        )
        assert str(exc_info.value) == error_message


def test_create_dataclass_subclass():
    for cls in [QuamBase, QuamRoot, QuamComponent]:

        @quam_dataclass
        class TestClass(cls):
            pass

        test_class = TestClass()
        assert isinstance(test_class, cls)
        assert is_dataclass(test_class)


@quam_dataclass
class BareQuamRoot(QuamRoot):
    pass


@quam_dataclass
class BareQuamComponent(QuamComponent):
    pass


@quam_dataclass(eq=False)
class QuamTest(QuamRoot):
    int_val: int
    quam_elem: QuamComponent
    quam_elem_list: List[QuamComponent]


def test_iterate_empty_quam_root():
    quam_root = BareQuamRoot()
    elems = list(quam_root.iterate_components())
    assert len(elems) == 0


def test_iterate_quam_root():
    quam_root = QuamTest(int_val=42, quam_elem=BareQuamComponent(), quam_elem_list=[])
    elems = list(quam_root.iterate_components())
    assert elems == [quam_root.quam_elem]

    quam_root.quam_elem_list.append(BareQuamComponent())
    elems = list(quam_root.iterate_components())
    assert elems == [quam_root.quam_elem, quam_root.quam_elem_list[0]]


def test_iterate_empty_quam_component():
    elem = BareQuamComponent()
    elems = list(elem.iterate_components())
    assert len(elems) == 1
    assert elems[0] is elem
    assert set(elem.get_attrs()) == set()


def test_iterate_quam_component_nested():
    elem = QuamTest(
        int_val=42, quam_elem=BareQuamComponent(), quam_elem_list=[BareQuamComponent()]
    )
    elems = list(elem.iterate_components())
    assert len(elems) == 2
    assert elems[0] is elem.quam_elem
    assert elems[1] is elem.quam_elem_list[0]

    assert set(elem.get_attrs()) == {"int_val", "quam_elem", "quam_elem_list"}


def test_iterate_quam_component_duplicate():
    @quam_dataclass
    class QuamTest(QuamRoot):
        quam_elem1: QuamComponent
        quam_elem2: QuamComponent

    components = [BareQuamComponent(), BareQuamComponent()]
    elem = QuamTest(quam_elem1=components[0], quam_elem2=components[1])
    elems = list(elem.iterate_components())
    assert elems == components

    quam_component = BareQuamComponent()
    elem = QuamTest(quam_elem1=quam_component, quam_elem2=quam_component)
    elems = list(elem.iterate_components())
    assert elems == [quam_component]


def test_iterate_quam_component_list_duplicate():
    @quam_dataclass
    class QuamTest(QuamRoot):
        quam_list: List[QuamComponent]

    components = [BareQuamComponent(), BareQuamComponent()]
    elem = QuamTest(quam_list=components)
    elems = list(elem.iterate_components())
    assert elems == components

    quam_component = BareQuamComponent()
    elem = QuamTest(quam_list=[quam_component, quam_component])
    elems = list(elem.iterate_components())
    assert elems == [quam_component]


def test_iterate_quam_with_elements():
    test_quam = QuamTest(
        int_val=42,
        quam_elem=BareQuamComponent(),
        quam_elem_list=[BareQuamComponent(), BareQuamComponent()],
    )

    elems = list(test_quam.iterate_components())
    assert len(elems) == 3
    assert all(isinstance(elem, QuamComponent) for elem in elems)


@quam_dataclass
class QuamComponentTest(QuamComponent):
    int_val: int
    quam_elem: QuamComponent
    quam_elem_list: List[QuamComponent]


@quam_dataclass
class NestedQuamTest(QuamRoot):
    int_val: int
    quam_elem: QuamComponentTest
    quam_elem_list: List[QuamComponentTest]


def test_iterate_components_nested():
    quam_component = QuamComponentTest(
        int_val=42,
        quam_elem=BareQuamComponent(),
        quam_elem_list=[BareQuamComponent(), BareQuamComponent()],
    )

    with pytest.raises(AttributeError):
        test_quam = NestedQuamTest(
            int_val=42,
            quam_elem=quam_component,
            quam_elem_list=[quam_component, quam_component],
        )

    quam_component.parent = None
    quam_component2 = QuamComponentTest(
        int_val=42,
        quam_elem=BareQuamComponent(),
        quam_elem_list=[BareQuamComponent(), BareQuamComponent()],
    )
    test_quam = NestedQuamTest(
        int_val=42,
        quam_elem=quam_component,
        quam_elem_list=[quam_component2, quam_component2],
    )

    elems = list(test_quam.iterate_components())
    assert len(elems) == 8
    assert all(isinstance(elem, QuamComponent) for elem in elems)


def test_quam_dict_element():
    elem = QuamDict(a=42)
    assert not isinstance(elem, QuamComponent)
    assert isinstance(elem, QuamBase)
    assert is_dataclass(elem)
    assert elem.a == 42
    assert elem.data == {"a": 42}

    elem.a = 43
    assert elem.a == 43
    assert elem.data == {"a": 43}

    elem.b = 44
    assert elem.b == 44
    assert elem.data == {"a": 43, "b": 44}

    elem["c"] = 45
    assert elem.c == 45
    assert elem.data == {"a": 43, "b": 44, "c": 45}


def test_iterate_components_dict():
    elem = BareQuamComponent()
    elem_dict = QuamDict(a=42, b=elem)

    elems = list(elem_dict.iterate_components())

    assert len(elems) == 1
    assert elems[0] is elem_dict.b


def test_nested_quam_dict_explicit():
    elem = QuamDict(subdict=QuamDict(a=42))

    assert elem.subdict.a == 42
    assert elem.subdict.data == {"a": 42}
    assert isinstance(elem.subdict, QuamDict)


def test_nested_quam_dict():
    elem = QuamDict(subdict=dict(a=42))

    assert elem.subdict.a == 42
    assert elem.subdict.data == {"a": 42}
    assert isinstance(elem.subdict, QuamDict)

    assert elem.to_dict() == {"subdict": {"a": 42}}


def test_quam_setattr_quam_dict_component():
    @quam_dataclass
    class TestQuamRoot(QuamRoot):
        quam_dict: dict = field(default_factory=dict)

    quam = TestQuamRoot()
    assert isinstance(quam.quam_dict, QuamDict)

    quam.quam_dict = {"a": 42}
    assert isinstance(quam.quam_dict, QuamDict)
    assert quam.quam_dict.a == 42

    quam.quam_dict = {"a": {"b": 43}}
    assert isinstance(quam.quam_dict, QuamDict)
    assert isinstance(quam.quam_dict.a, QuamDict)
    assert quam.quam_dict.a.b == 43


def test_parent_attr_name(BareQuamComponent):
    inner_quam_component = BareQuamComponent()
    assert inner_quam_component.parent is None
    with pytest.raises(AttributeError):
        inner_quam_component.parent.get_attr_name(inner_quam_component)

    @quam_dataclass
    class OuterQuamComponent(QuamComponent):
        quam_component: QuamComponent

    outer_quam_component = OuterQuamComponent(quam_component=inner_quam_component)
    assert outer_quam_component.quam_component is inner_quam_component
    assert inner_quam_component.parent is outer_quam_component
    assert (
        inner_quam_component.parent.get_attr_name(inner_quam_component)
        == "quam_component"
    )

    outer_quam_component.quam_component = None
    with pytest.raises(AttributeError):
        inner_quam_component.parent.get_attr_name(inner_quam_component)


def test_get_attr_name_with_reference():
    @quam_dataclass
    class TestQuamComponent(QuamComponent):
        a: dict
        b: dict

    component = TestQuamComponent(a="#./b", b={"value": 43})
    
    assert component.a is component.b
    assert component.get_attr_name(component.b) == "b"


def test_quam_dict_get_attr_name_with_reference():
    d = QuamDict(a="#./b", b={"value": 42})

    assert d.a is d.b
    assert d.get_attr_name(d.b) == "b"



import click

from quam.config.cli import config_command, migrate_command


@click.group()
def cli() -> None:
    pass


cli.add_command(config_command)
cli.add_command(migrate_command)


def main() -> None:
    cli()



import pytest
import json

from quam.components.ports.analog_outputs import MWFEMAnalogOutputPort
from quam.components.ports.analog_inputs import MWFEMAnalogInputPort


def test_mw_fem_analog_output_port():
    with pytest.raises(TypeError):
        MWFEMAnalogOutputPort()

    with pytest.raises(TypeError):
        port = MWFEMAnalogOutputPort("con1", 1, 2)

    with pytest.raises(ValueError):
        port = MWFEMAnalogOutputPort("con1", 1, 2, band=1)

    port = MWFEMAnalogOutputPort("con1", 1, 2, band=1, upconverter_frequency=5e9)
    assert port.controller_id == "con1"
    assert port.fem_id == 1
    assert port.port_id == 2
    assert port.port_tuple == ("con1", 1, 2)
    assert port.port_type == "analog_output"
    assert port.band == 1
    assert port.upconverter_frequency == 5e9
    assert port.upconverters is None
    assert port.delay == 0
    assert port.shareable == False
    assert port.sampling_rate == 1e9
    assert port.full_scale_power_dbm == -11

    assert port.to_dict(include_defaults=False) == {
        "__class__": "quam.components.ports.analog_outputs.MWFEMAnalogOutputPort",
        "controller_id": "con1",
        "fem_id": 1,
        "port_id": 2,
        "band": 1,
        "upconverter_frequency": 5e9,
    }

    assert port.get_port_properties() == {
        "band": 1,
        "delay": 0,
        "shareable": False,
        "sampling_rate": 1e9,
        "full_scale_power_dbm": -11,
        "upconverter_frequency": 5e9,
    }

    cfg = {"controllers": {}}
    port.apply_to_config(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "fems": {
                    1: {
                        "type": "MW",
                        "analog_outputs": {
                            2: {
                                "band": 1,
                                "upconverter_frequency": 5e9,
                                "delay": 0,
                                "shareable": False,
                                "sampling_rate": 1e9,
                                "full_scale_power_dbm": -11,
                            }
                        },
                    }
                }
            }
        }
    }


def test_mw_fem_analog_input_ports():
    with pytest.raises(TypeError):
        MWFEMAnalogInputPort()

    with pytest.raises(TypeError):
        port = MWFEMAnalogInputPort("con1", 1, 2)

    port = MWFEMAnalogInputPort("con1", 1, 2, band=1, downconverter_frequency=5e9)

    assert port.controller_id == "con1"
    assert port.fem_id == 1
    assert port.port_id == 2
    assert port.port_tuple == ("con1", 1, 2)
    assert port.port_type == "analog_input"
    assert port.band == 1
    assert port.downconverter_frequency == 5e9
    assert port.sampling_rate == 1e9
    assert port.shareable == False

    assert port.to_dict(include_defaults=False) == {
        "__class__": "quam.components.ports.analog_inputs.MWFEMAnalogInputPort",
        "controller_id": "con1",
        "fem_id": 1,
        "port_id": 2,
        "band": 1,
        "downconverter_frequency": 5e9,
    }

    assert port.get_port_properties() == {
        "band": 1,
        "downconverter_frequency": 5e9,
        "sampling_rate": 1e9,
        "shareable": False,
    }

    cfg = {"controllers": {}}
    port.apply_to_config(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "fems": {
                    1: {
                        "type": "MW",
                        "analog_inputs": {
                            2: {
                                "band": 1,
                                "downconverter_frequency": 5e9,
                                "sampling_rate": 1e9,
                                "shareable": False,
                            }
                        },
                    }
                }
            }
        }
    }

    port.gain_db = 10
    assert port.get_port_properties() == {
        "band": 1,
        "downconverter_frequency": 5e9,
        "sampling_rate": 1e9,
        "shareable": False,
        "gain_db": 10,
    }


def test_mw_fem_upconverter_config_generation():
    port = MWFEMAnalogOutputPort(
        "con1",
        1,
        2,
        band=1,
        upconverters={1: {"frequency": 4.5e9}, 2: {"frequency": 5.5e9}},
    )
    cfg = {"controllers": {}}
    port.apply_to_config(cfg)

    assert cfg["controllers"]["con1"]["fems"][1]["analog_outputs"][2] == {
        "band": 1,
        "delay": 0,
        "shareable": False,
        "sampling_rate": 1000000000.0,
        "full_scale_power_dbm": -11,
        "upconverters": {
            1: {"frequency": 4500000000.0},
            2: {"frequency": 5500000000.0},
        },
    }
    json.dumps(cfg, indent=4)
